# mypy: disable-error-code=method-assign

import unittest
from pathlib import Path
from subprocess import CalledProcessError
from unittest.mock import Mock, patch

from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.outdefs import ListStrTemplateOutputDefinition
from jupyter_deploy.engine.terraform.tf_down import TerraformDownHandler


class TestTerraformDownHandler(unittest.TestCase):
    def get_mock_outputs_handler_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return the mock outputs handler."""
        mock_handler = Mock()
        mock_get_declared_output_def = Mock()
        mock_handler.get_declared_output_def = mock_get_declared_output_def

        mock_get_declared_output_def.return_value = ListStrTemplateOutputDefinition(
            output_name="persisting_resources", value=[]
        )
        return mock_handler, {"get_declared_output_def": mock_get_declared_output_def}

    def get_mock_manifest_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return mock manifest with functions defined as mock."""
        mock_manifest = Mock()
        mock_get_engine = Mock()
        mock_get_engine.return_value = EngineType.TERRAFORM
        mock_manifest.get_engine = mock_get_engine
        return mock_manifest, {"get_engine": mock_get_engine}

    def test_init_sets_attributes(self) -> None:
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)

        self.assertEqual(handler.project_path, project_path)
        self.assertEqual(handler.project_manifest, mock_manifest)
        self.assertEqual(handler.engine, EngineType.TERRAFORM)

    @patch("jupyter_deploy.engine.terraform.tf_down.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.rich_console")
    def test_destroy_success_no_persisting_resources(self, mock_console: Mock, mock_cmd_utils: Mock) -> None:
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        engine_path = project_path / "engine"
        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (0, False)

        handler.destroy()

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once_with(
            ["terraform", "destroy"], exec_dir=engine_path
        )
        mock_console_instance.print.assert_called_once()
        self.assertTrue(mock_console_instance.print.call_args[0][0].lower().find("success") >= 0)

    @patch("jupyter_deploy.engine.terraform.tf_down.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.rich_console")
    def test_destroy_handles_error(self, mock_console: Mock, mock_cmd_utils: Mock) -> None:
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        engine_path = project_path / "engine"
        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (1, False)

        handler.destroy()

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once_with(
            ["terraform", "destroy"], exec_dir=engine_path
        )
        mock_console_instance.print.assert_called_once()
        self.assertTrue(mock_console_instance.print.call_args[0][0].lower().find("error") >= 0)

    @patch("jupyter_deploy.engine.terraform.tf_down.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.rich_console")
    def test_destroy_handles_timeout(self, mock_console: Mock, mock_cmd_utils: Mock) -> None:
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        engine_path = project_path / "engine"

        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (0, True)

        handler.destroy()

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once_with(
            ["terraform", "destroy"], exec_dir=engine_path
        )
        mock_console_instance.print.assert_called_once()
        self.assertTrue(mock_console_instance.print.call_args[0][0].lower().find("error") >= 0)

    @patch("jupyter_deploy.engine.terraform.tf_down.cmd_utils")
    def test_destroy_propagates_exceptions(self, mock_cmd_utils: Mock) -> None:
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.side_effect = Exception("Command failed")

        with self.assertRaises(Exception) as context:
            handler.destroy()

        self.assertEqual(str(context.exception), "Command failed")
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once()

    @patch("jupyter_deploy.engine.terraform.tf_down.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.rich_console")
    def test_destroy_with_auto_approve(self, mock_console: Mock, mock_cmd_utils: Mock) -> None:
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (0, False)

        handler.destroy(auto_approve=True)

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once()
        cmd_args = mock_cmd_utils.run_cmd_and_pipe_to_terminal.call_args[0][0]
        self.assertIn("-auto-approve", cmd_args)

    @patch("jupyter_deploy.engine.terraform.tf_down.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.rich_console")
    def test_destroy_with_persisting_resources_runs_dryrun_and_stop_without_yes_flag(
        self, mock_console: Mock, mock_cmd_utils: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)

        # Mock the get_persisting_resources method to return resources
        persisting_resources = [
            'aws_ebs_volume.additional_volumes["0"]',
            'aws_efs_file_system.additional_file_systems["0"]',
        ]
        handler.get_persisting_resources = Mock(return_value=persisting_resources)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        # Mock successful dry-run
        mock_cmd_utils.run_cmd_and_capture_output.return_value = "dry-run output"

        # Act
        handler.destroy(auto_approve=False)

        # Assert
        mock_cmd_utils.run_cmd_and_capture_output.assert_called_once()
        cmd_args = mock_cmd_utils.run_cmd_and_capture_output.call_args[0][0]
        self.assertIn("--dry-run", cmd_args)
        for resource in persisting_resources:
            self.assertIn(resource, cmd_args)

        # Verify we don't proceed with terraform destroy
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_not_called()

    @patch("jupyter_deploy.engine.terraform.tf_down.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.rich_console")
    def test_destroy_remove_persisting_resources_and_calls_destroy_happy_path(
        self, mock_console: Mock, mock_cmd_utils: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)

        # Mock the get_persisting_resources method to return resources
        persisting_resources = [
            'aws_ebs_volume.additional_volumes["0"]',
            'aws_efs_file_system.additional_file_systems["0"]',
        ]
        handler.get_persisting_resources = Mock(return_value=persisting_resources)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        # Mock successful dry-run
        mock_cmd_utils.run_cmd_and_capture_output.return_value = "dry-run output"

        # Mock successful state removal and destroy
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (0, False)

        # Act
        handler.destroy(auto_approve=True)

        # Assert
        # Check dry-run call
        mock_cmd_utils.run_cmd_and_capture_output.assert_called_once()
        dryrun_args = mock_cmd_utils.run_cmd_and_capture_output.call_args[0][0]
        self.assertIn("--dry-run", dryrun_args)

        # Check actual state removal call
        self.assertEqual(2, mock_cmd_utils.run_cmd_and_pipe_to_terminal.call_count)
        rm_call_args = mock_cmd_utils.run_cmd_and_pipe_to_terminal.call_args_list[0][0][0]

        self.assertNotIn("--dry-run", rm_call_args)
        for resource in persisting_resources:
            self.assertIn(resource, rm_call_args)

        # Check destroy call with auto-approve
        destroy_call_args = mock_cmd_utils.run_cmd_and_pipe_to_terminal.call_args_list[1][0][0]
        self.assertEqual(["terraform", "destroy", "-auto-approve"], destroy_call_args)

    @patch("jupyter_deploy.engine.terraform.tf_down.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.rich_console")
    def test_destroy_raises_on_failed_dryrun(self, mock_console: Mock, mock_cmd_utils: Mock) -> None:
        # Setup
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)

        # Mock the get_persisting_resources method to return resources
        persisting_resources = [
            'aws_ebs_volume.additional_volumes["0"]',
            'aws_efs_file_system.additional_file_systems["0"]',
        ]
        handler.get_persisting_resources = Mock(return_value=persisting_resources)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        # Mock failed dry-run
        error_msg = "Some terraform error"
        mock_cmd_utils.run_cmd_and_capture_output.side_effect = CalledProcessError(1, "cmd", stderr=error_msg.encode())

        # Act
        handler.destroy(auto_approve=True)

        # Assert
        mock_cmd_utils.run_cmd_and_capture_output.assert_called_once()

        # Verify we don't proceed with terraform state rm or destroy
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_not_called()

    @patch("jupyter_deploy.engine.terraform.tf_down.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.rich_console")
    def test_destroy_raises_on_failed_remove_persisting_resources_without_destroying(
        self, mock_console: Mock, mock_cmd_utils: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)

        # Mock the get_persisting_resources method to return resources
        persisting_resources = [
            'aws_ebs_volume.additional_volumes["0"]',
            'aws_efs_file_system.additional_file_systems["0"]',
        ]
        handler.get_persisting_resources = Mock(return_value=persisting_resources)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        # Mock successful dry-run
        mock_cmd_utils.run_cmd_and_capture_output.return_value = "dry-run output"

        # Mock failed state removal
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (1, False)

        # Act
        handler.destroy(auto_approve=True)

        # Assert
        # Check dry-run call
        mock_cmd_utils.run_cmd_and_capture_output.assert_called_once()

        # Check actual state removal call
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once()

        # Verify we don't proceed with terraform destroy
        self.assertEqual(1, mock_cmd_utils.run_cmd_and_pipe_to_terminal.call_count)

    @patch("jupyter_deploy.engine.terraform.tf_down.fs_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.rich_console")
    def test_passes_destroy_tfvars_file_when_available(
        self, mock_console: Mock, mock_cmd_utils: Mock, mock_fs_utils: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)
        engine_path = project_path / "engine"

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        # Mock fs_utils to indicate that destroy.tfvars exists
        mock_fs_utils.file_exists.return_value = True

        # Mock successful command execution
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (0, False)

        # Define expected tfvars file path (using tf_constants.TF_DESTROY_PRESET_FILENAME)
        destroy_tfvars_path = engine_path / "presets" / "destroy.tfvars"

        # Act
        handler.destroy(auto_approve=True)

        # Assert
        mock_fs_utils.file_exists.assert_called_once()
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once()

        # Check that the var-file option was included in the command
        cmd_args = mock_cmd_utils.run_cmd_and_pipe_to_terminal.call_args[0][0]
        var_file_arg = f"-var-file={destroy_tfvars_path.absolute()}"
        self.assertIn(var_file_arg, cmd_args)

    @patch("jupyter_deploy.engine.terraform.tf_down.fs_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_down.rich_console")
    def test_skips_passing_destroy_tfvars_file_when_unavailable(
        self, mock_console: Mock, mock_cmd_utils: Mock, mock_fs_utils: Mock
    ) -> None:
        # Setup
        project_path = Path("/mock/project")
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        handler = TerraformDownHandler(project_path=project_path, project_manifest=mock_manifest)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        # Mock fs_utils to indicate that destroy.tfvars does NOT exist
        mock_fs_utils.file_exists.return_value = False

        # Mock successful command execution
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (0, False)

        # Act
        handler.destroy(auto_approve=True)

        # Assert
        mock_fs_utils.file_exists.assert_called_once()
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once()

        # Check that the var-file option was NOT included in the command
        cmd_args = mock_cmd_utils.run_cmd_and_pipe_to_terminal.call_args[0][0]
        for arg in cmd_args:
            self.assertFalse(arg.startswith("-var-file="))



from rich import console as rich_console

from jupyter_deploy.engine.engine_config import EngineConfigHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform import tf_config
from jupyter_deploy.engine.vardefs import TemplateVariableDefinition
from jupyter_deploy.handlers.base_project_handler import BaseProjectHandler


class ConfigHandler(BaseProjectHandler):
    _handler: EngineConfigHandler

    def __init__(self, output_filename: str | None = None) -> None:
        """Base class to manage the configuration of a jupyter-deploy project."""
        super().__init__()
        self.preset_name: str | None = None

        if self.engine == EngineType.TERRAFORM:
            self._handler = tf_config.TerraformConfigHandler(
                project_path=self.project_path,
                project_manifest=self.project_manifest,
                output_filename=output_filename,
            )
        else:
            raise NotImplementedError(f"ConfigHandler implementation not found for engine: {self.engine}")

    def validate_and_set_preset(self, preset_name: str | None, will_reset_variables: bool = False) -> bool:
        """Return True if the settings are correct."""
        console = rich_console.Console()

        # first, verify whether there are recorded variables values from user inputs
        # if yes, do NOT use the preset defaults.
        # `jd config` records values automatically, and we want users to be able to rerun `jd config`
        # without getting prompted again or having their previous choices overridden by defaults.
        if not will_reset_variables and self._handler.has_recorded_variables():
            console.rule()
            console.print(
                ":magnifying_glass_tilted_right: Detected variables values that [bold]jupyter-deploy[/] "
                "recorded previously."
            )
            console.print("Recorded values take precedent over any default preset.")
            console.print("You can override any recorded variable value with [bold cyan]--variable-name <value>[/].")
            preset_name = None

        preset_valid = preset_name is None or self._handler.verify_preset_exists(preset_name)

        if not preset_valid:
            valid_presets = self._handler.list_presets()
            console.print(f":x: preset [bold]{preset_name}[/] is invalid for this template.", style="red")
            console.print(f"Valid presets: {valid_presets}")

        # then set the preset, which may have been overridden to None if it detected recorded values.
        self.preset_name = preset_name
        return preset_valid

    def has_used_preset(self, expected_preset_name: str | None) -> bool:
        """Return True if the handler has used the preset.

        Always return True if expected_preset argument is None.
        This method returns False when it detected recorded variables values that the users
        decided not to reset.
        """
        return expected_preset_name is None or expected_preset_name == self.preset_name

    def reset_recorded_variables(self) -> None:
        """Delete the file in the project dir where the previous inputs were recorded."""
        self._handler.reset_recorded_variables()

    def reset_recorded_secrets(self) -> None:
        """Delete the file in the project dir where the secrets were recorded."""
        self._handler.reset_recorded_secrets()

    def verify_requirements(self) -> bool:
        """Return True if the user has installed all the required dependencies."""
        return self._handler.verify_requirements()

    def configure(self, variable_overrides: dict[str, TemplateVariableDefinition] | None = None) -> bool:
        """Main method to set the inputs for the project, return True on success"""
        return self._handler.configure(preset_name=self.preset_name, variable_overrides=variable_overrides)

    def record(self, record_vars: bool = False, record_secrets: bool = False) -> None:
        """Save the values of the variables to disk in the project dir."""
        self._handler.record(record_vars=record_vars, record_secrets=record_secrets)



import subprocess
import unittest
from collections import OrderedDict
from pathlib import Path
from unittest.mock import Mock, patch

from pydantic import ValidationError

from jupyter_deploy.engine.terraform.tf_config import TerraformConfigHandler
from jupyter_deploy.engine.terraform.tf_constants import (
    TF_DEFAULT_PLAN_FILENAME,
    TF_ENGINE_DIR,
    TF_PRESETS_DIR,
)
from jupyter_deploy.engine.vardefs import (
    BoolTemplateVariableDefinition,
    DictStrTemplateVariableDefinition,
    FloatTemplateVariableDefinition,
    ListStrTemplateVariableDefinition,
    StrTemplateVariableDefinition,
    TemplateVariableDefinition,
)


class TestTerraformConfigHandler(unittest.TestCase):
    MOCK_OVERRIDE_PRESET_PATH = Path("/mock/path/engine/jdinputs.preset.override.tfvars")
    MOCK_RECORD_VARS_PATH = Path("/mock/path/engine/jdinputs.auto.tfvars")
    MOCK_RECORD_SECRETS_PATH = Path("/mock/path/engine/jdinputs.secrets.auto.tfvars")

    def get_mock_variable_handler_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        mock_handler = Mock()
        mock_get_recorded_variables_filepath = Mock()
        mock_get_recorded_secrets_filepath = Mock()
        mock_reset_recorded_variables = Mock()
        mock_reset_recorded_secrets = Mock()
        mock_sync_engine_varfiles = Mock()
        mock_sync_variables_config = Mock()
        mock_get_template_variables = Mock()
        mock_update_variable_records = Mock()
        mock_create_filtered_preset_file = Mock()

        mock_handler.get_recorded_variables_filepath = mock_get_recorded_variables_filepath
        mock_handler.get_recorded_secrets_filepath = mock_get_recorded_secrets_filepath
        mock_handler.reset_recorded_variables = mock_reset_recorded_variables
        mock_handler.reset_recorded_secrets = mock_reset_recorded_secrets
        mock_handler.sync_engine_varfiles_with_project_variables_config = mock_sync_engine_varfiles
        mock_handler.sync_project_variables_config = mock_sync_variables_config
        mock_handler.get_template_variables = mock_get_template_variables
        mock_handler.update_variable_records = mock_update_variable_records
        mock_handler.create_filtered_preset_file = mock_create_filtered_preset_file

        mock_get_recorded_variables_filepath.return_value = TestTerraformConfigHandler.MOCK_RECORD_VARS_PATH
        mock_get_recorded_secrets_filepath.return_value = TestTerraformConfigHandler.MOCK_RECORD_SECRETS_PATH
        mock_create_filtered_preset_file.return_value = TestTerraformConfigHandler.MOCK_OVERRIDE_PRESET_PATH

        return mock_handler, {
            "get_recorded_variables_filepath": mock_get_recorded_variables_filepath,
            "get_recorded_secrets_filepath": mock_get_recorded_secrets_filepath,
            "reset_recorded_variables": mock_reset_recorded_variables,
            "reset_recorded_secrets": mock_reset_recorded_secrets,
            "sync_engine_varfiles_with_project_variables_config": mock_sync_engine_varfiles,
            "sync_project_variables_config": mock_sync_variables_config,
            "get_template_variables": mock_get_template_variables,
            "update_variables_record": mock_update_variable_records,
            "create_filtered_preset_file": mock_create_filtered_preset_file,
        }

    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_class_can_instantiate(self, mock_variable_handler_cls: Mock) -> None:
        # Arrange
        path = Path("/fake/path")
        manifest = Mock()

        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        handler = TerraformConfigHandler(path, manifest)

        # Assert
        self.assertIsNotNone(handler)
        self.assertEqual(handler.plan_out_path, path / TF_ENGINE_DIR / TF_DEFAULT_PLAN_FILENAME)
        self.assertEqual(handler.project_manifest, manifest)
        self.assertEqual(handler.tf_variables_handler, mock_vars_handler)

        # expensive methods of EngineVariablesHandler are not called
        mock_vars_fns["get_template_variables"].assert_not_called()
        mock_vars_fns["update_variables_record"].assert_not_called()

    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_class_uses_custom_output_file_when_provided(self, mock_variable_handler_cls: Mock) -> None:
        # Arrange
        path = Path("/fake/path")
        manifest = Mock()
        custom_output = "custom-output-file"
        mock_vars_handler, _ = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        handler = TerraformConfigHandler(path, manifest, output_filename=custom_output)

        # Assert
        self.assertIsNotNone(handler)
        self.assertEqual(handler.plan_out_path, path / TF_ENGINE_DIR / custom_output)
        self.assertEqual(handler.tf_variables_handler, mock_vars_handler)

    @patch("jupyter_deploy.fs_utils.file_exists")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_verify_preset_exists_calls_fs_util(self, mock_variable_handler_cls: Mock, mock_file_exists: Mock) -> None:
        # Arrange
        path = Path("/fake/path")
        mock_vars_handler, _ = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        handler = TerraformConfigHandler(path, Mock())

        # Act
        handler.verify_preset_exists("all")

        # Assert
        mock_file_exists.assert_called_once_with(
            file_path=path / TF_ENGINE_DIR / TF_PRESETS_DIR / "defaults-all.tfvars"
        )

    @patch("jupyter_deploy.fs_utils.find_matching_filenames")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_list_presets_calls_fs_util(self, mock_variable_handler_cls: Mock, mock_find: Mock) -> None:
        mock_find.return_value = [
            "defaults-all.tfvars",
            "defaults-base.tfvars",
            "defaults-all-except-instance.tfvars",
        ]

        # Arrange
        path = Path("/fake/path")
        mock_vars_handler, _ = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        handler = TerraformConfigHandler(path, Mock())

        # Act
        presets = handler.list_presets()

        # Assert
        self.assertEqual(sorted(["all", "all-except-instance", "base", "none"]), sorted(presets))
        mock_find.assert_called_once_with(
            dir_path=handler.engine_dir_path / TF_PRESETS_DIR,
            file_pattern="defaults-*.tfvars",
        )

    @patch("jupyter_deploy.fs_utils.find_matching_filenames")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_list_presets_always_returns_none(self, mock_variable_handler_cls: Mock, mock_find: Mock) -> None:
        mock_find.return_value = []

        # Arrange
        path = Path("/fake/path")
        mock_vars_handler, _ = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        handler = TerraformConfigHandler(path, Mock())

        # Act
        presets = handler.list_presets()

        # Assert
        self.assertEqual(["none"], presets)

    @patch("jupyter_deploy.verify_utils.verify_tools_installation")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_verify_requirements_pulls_manifest_and_call_verify(
        self, mock_variable_handler_cls: Mock, mock_verify: Mock
    ) -> None:
        # Arrange
        path = Path("/fake/path")
        mock_manifest = Mock()
        mock_get_requirements = Mock()
        mock_manifest.get_requirements = mock_get_requirements

        mock_req1 = Mock()
        mock_req2 = Mock()
        mock_get_requirements.return_value = [mock_req1, mock_req2]

        mock_verify.return_value = True
        mock_vars_handler, _ = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler

        handler = TerraformConfigHandler(path, mock_manifest)

        # Act
        result = handler.verify_requirements()

        # Assert
        self.assertTrue(result)
        mock_get_requirements.assert_called_once()
        mock_verify.assert_called_once_with([mock_req1, mock_req2])

    @patch("jupyter_deploy.verify_utils.verify_tools_installation")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_verify_requirements_raises_when_checks_raise(
        self, mock_variable_handler_cls: Mock, mock_verify: Mock
    ) -> None:
        # Arrange
        mock_manifest = Mock()
        mock_get_requirements = Mock()
        mock_manifest.get_requirements = mock_get_requirements
        mock_verify.side_effect = Exception("Terraform check failed")

        mock_vars_handler, _ = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler

        handler = TerraformConfigHandler(Path("/fake/path"), mock_manifest)

        # Act & Assert
        with self.assertRaises(Exception) as e:
            handler.verify_requirements()
        self.assertEqual(str(e.exception), "Terraform check failed")

    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_reset_recorded_variables_calls_vars_handler(self, mock_variable_handler_cls: Mock) -> None:
        # Arrange
        path = Path("/fake/path")
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        handler = TerraformConfigHandler(path, Mock())

        # Act
        handler.reset_recorded_variables()

        # Assert
        mock_vars_fns["reset_recorded_variables"].assert_called_once()

    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_reset_recorded_secrets_calls_vars_handler(self, mock_variable_handler_cls: Mock) -> None:
        # Arrange
        path = Path("/fake/path")
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        handler = TerraformConfigHandler(path, Mock())

        # Act
        handler.reset_recorded_secrets()

        # Assert
        mock_vars_fns["reset_recorded_secrets"].assert_called_once()

    @patch("jupyter_deploy.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_configure_calls_tf_init(self, mock_variable_handler_cls: Mock, mock_run_cmd: Mock) -> None:
        # Arrange
        mock_vars_handler, _ = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        mock_run_cmd.return_value = (0, False)  # Return code 0, no timeout
        handler = TerraformConfigHandler(Path("/fake/path"), Mock())

        # Mock the variables handler
        mock_variables_handler = Mock()
        handler.variables_handler = mock_variables_handler

        # Act
        handler.configure()

        # Assert
        self.assertGreaterEqual(mock_run_cmd.call_count, 1)
        init_cmds = mock_run_cmd.mock_calls[0][1][0]
        init_kwargs = mock_run_cmd.mock_calls[0][2]
        self.assertEqual(init_cmds[:2], ["terraform", "init"])
        self.assertEqual(init_kwargs, {"exec_dir": Path("/fake/path/engine")})

    @patch("jupyter_deploy.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_configure_calls_tf_plan_with_a_named_plan(
        self, mock_variable_handler_cls: Mock, mock_run_cmd: Mock
    ) -> None:
        # Arrange
        # First call for init returns success
        # Second call for plan returns success
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        mock_run_cmd.side_effect = [(0, False), (0, False)]
        handler = TerraformConfigHandler(Path("/fake/path"), Mock())

        # Act
        result = handler.configure()

        # Assert
        # Check the second call was to plan
        self.assertTrue(result)
        self.assertEqual(mock_run_cmd.call_count, 2)

        plan_cmds = mock_run_cmd.mock_calls[1][1][0]
        plan_kwargs = mock_run_cmd.mock_calls[0][2]
        self.assertEqual(len(plan_cmds), 3)
        self.assertEqual(plan_cmds[:2], ["terraform", "plan"])
        self.assertIn("-out=", plan_cmds[2])
        self.assertEqual(plan_kwargs, {"exec_dir": Path("/fake/path/engine")})
        mock_vars_fns["sync_engine_varfiles_with_project_variables_config"].assert_called_once()

    @patch("jupyter_deploy.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_configure_calls_tf_plan_passes_preset(self, mock_variable_handler_cls: Mock, mock_run_cmd: Mock) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        # First call for init returns success
        # Second call for plan returns success
        mock_run_cmd.side_effect = [(0, False), (0, False)]
        path = Path("/fake/path")
        handler = TerraformConfigHandler(path, Mock())

        # Act
        result = handler.configure(preset_name="all")

        # Assert
        # Check the second call was to plan
        self.assertTrue(result)
        self.assertEqual(mock_run_cmd.call_count, 2)

        plan_cmds = mock_run_cmd.mock_calls[1][1][0]
        plan_kwargs = mock_run_cmd.mock_calls[0][2]
        self.assertEqual(len(plan_cmds), 4)
        self.assertEqual(plan_cmds[:2], ["terraform", "plan"])
        self.assertIn("-out=", plan_cmds[2])
        self.assertEqual(plan_kwargs, {"exec_dir": Path("/fake/path/engine")})

        expect_called_path = path / TF_ENGINE_DIR / TF_PRESETS_DIR / "defaults-all.tfvars"
        expect_path = TestTerraformConfigHandler.MOCK_OVERRIDE_PRESET_PATH
        self.assertEqual(f"-var-file={expect_path.absolute()}", plan_cmds[3])
        mock_vars_fns["sync_engine_varfiles_with_project_variables_config"].assert_called_once()
        mock_vars_fns["create_filtered_preset_file"].assert_called_once_with(expect_called_path)

    @patch("jupyter_deploy.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_configure_calls_tf_plan_with_variable_override(
        self, mock_variable_handler_cls: Mock, mock_run_cmd: Mock
    ) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        # First call for init returns success
        # Second call for plan returns success
        mock_run_cmd.side_effect = [(0, False), (0, False)]
        path = Path("/fake/path")
        handler = TerraformConfigHandler(path, Mock())

        mock_var1 = Mock(spec=StrTemplateVariableDefinition)
        mock_var2 = Mock(spec=FloatTemplateVariableDefinition)
        mock_var3 = Mock(spec=BoolTemplateVariableDefinition)
        mock_var4 = Mock(spec=ListStrTemplateVariableDefinition)
        mock_var5 = Mock(spec=DictStrTemplateVariableDefinition)
        mock_variables: dict[str, TemplateVariableDefinition] = OrderedDict(
            {"var1": mock_var1, "var2": mock_var2, "var3": mock_var3, "var4": mock_var4, "var5": mock_var5}
        )
        for idx, key in enumerate(mock_variables.keys()):
            mock_variables[key].variable_name = f"var{idx + 1}"

        mock_var1.assigned_value = "some-value"
        mock_var2.assigned_value = 3.1459
        mock_var3.assigned_value = True
        mock_var4.assigned_value = ["email1@example.com", "email2@example.com"]
        mock_var5.assigned_value = {"Key1": "Val1", "Key2": "Val2"}

        # Act
        result = handler.configure(preset_name="all", variable_overrides=mock_variables)

        # Assert
        # Check the second call was to plan
        self.assertTrue(result)
        self.assertEqual(mock_run_cmd.call_count, 2)
        mock_vars_fns["sync_engine_varfiles_with_project_variables_config"].assert_called_once()
        mock_vars_fns["create_filtered_preset_file"].assert_called_once()
        plan_cmds = mock_run_cmd.mock_calls[1][1][0]

        expect_path = TestTerraformConfigHandler.MOCK_OVERRIDE_PRESET_PATH
        self.assertEqual(f"-var-file={expect_path.absolute()}", plan_cmds[3])

        plan_cmds_len = len(plan_cmds)

        # should append the 3 variables as [-var, varname:varvalue]
        self.assertEqual("-var", plan_cmds[plan_cmds_len - 10])
        self.assertEqual("var1=some-value", plan_cmds[plan_cmds_len - 9])

        self.assertEqual("-var", plan_cmds[plan_cmds_len - 8])
        self.assertEqual("var2=3.1459", plan_cmds[plan_cmds_len - 7])

        self.assertEqual("-var", plan_cmds[plan_cmds_len - 6])
        self.assertEqual("var3=true", plan_cmds[plan_cmds_len - 5])

        self.assertEqual("-var", plan_cmds[plan_cmds_len - 4])
        self.assertEqual('var4=["email1@example.com", "email2@example.com"]', plan_cmds[plan_cmds_len - 3])

        self.assertEqual("-var", plan_cmds[plan_cmds_len - 2])
        self.assertEqual('var5={"Key1": "Val1", "Key2": "Val2"}', plan_cmds[plan_cmds_len - 1])

    @patch("jupyter_deploy.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("rich.console.Console")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_configure_does_not_call_plan_if_tf_init_fails(
        self, mock_variable_handler_cls: Mock, mock_console: Mock, mock_run_cmd: Mock
    ) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        mock_run_cmd.return_value = (1, False)  # Return code 1 (failure), no timeout
        mock_console_instance = Mock()
        mock_console.return_value = mock_console_instance
        handler = TerraformConfigHandler(Path("/fake/path"), Mock())

        # Act
        result = handler.configure()

        # Assert
        self.assertFalse(result)
        self.assertEqual(mock_run_cmd.call_count, 1)  # Only init should be called
        mock_cmd_call = mock_run_cmd.mock_calls[0]
        self.assertEqual(mock_cmd_call[1][0][:2], ["terraform", "init"])
        mock_vars_fns["sync_engine_varfiles_with_project_variables_config"].assert_not_called()

    @patch("jupyter_deploy.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("rich.console.Console")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_configure_does_not_call_plan_if_tf_init_timesout(
        self, mock_variable_handler_cls: Mock, mock_console: Mock, mock_run_cmd: Mock
    ) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        mock_run_cmd.return_value = (0, True)  # Return code 0, but timed out
        mock_console_instance = Mock()
        mock_console.return_value = mock_console_instance
        handler = TerraformConfigHandler(Path("/fake/path"), Mock())

        # Act
        result = handler.configure()

        # Assert
        self.assertFalse(result)
        self.assertEqual(mock_run_cmd.call_count, 1)  # Only init should be called
        mock_cmd_call = mock_run_cmd.mock_calls[0]
        self.assertEqual(mock_cmd_call[1][0][:2], ["terraform", "init"])
        mock_vars_fns["sync_engine_varfiles_with_project_variables_config"].assert_not_called()

    @patch("jupyter_deploy.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("rich.console.Console")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_configure_print_to_console_if_plan_fails(
        self, mock_variable_handler_cls: Mock, mock_console: Mock, mock_run_cmd: Mock
    ) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        # First call for init returns success
        # Second call for plan returns failure
        mock_run_cmd.side_effect = [(0, False), (1, False)]
        mock_console_instance = Mock()
        mock_console.return_value = mock_console_instance
        handler = TerraformConfigHandler(Path("/fake/path"), Mock())

        # Act
        result = handler.configure()

        # Assert
        self.assertFalse(result)
        self.assertEqual(mock_run_cmd.call_count, 2)
        mock_vars_fns["sync_engine_varfiles_with_project_variables_config"].assert_called_once()
        self.assertEqual(mock_console_instance.print.call_count, 1)
        mock_print_call = mock_console_instance.print.mock_calls[0]
        self.assertTrue(type(mock_print_call[1][0]) == str)  # noqa: E721
        self.assertTrue(len(mock_print_call[1][0]) > 0)

    @patch("jupyter_deploy.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("rich.console.Console")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_configure_print_to_console_if_plan_timesout(
        self, mock_variable_handler_cls: Mock, mock_console: Mock, mock_run_cmd: Mock
    ) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        # First call for init returns success
        # Second call for plan returns timeout
        mock_run_cmd.side_effect = [(0, False), (0, True)]
        mock_console_instance = Mock()
        mock_console.return_value = mock_console_instance
        handler = TerraformConfigHandler(Path("/fake/path"), Mock())

        # Act
        result = handler.configure()

        # Assert
        self.assertFalse(result)
        self.assertEqual(mock_run_cmd.call_count, 2)
        mock_vars_fns["sync_engine_varfiles_with_project_variables_config"].assert_called_once()
        self.assertEqual(mock_console_instance.print.call_count, 1)
        mock_print_call = mock_console_instance.print.mock_calls[0]
        self.assertTrue(type(mock_print_call[1][0]) == str)  # noqa: E721
        self.assertTrue(len(mock_print_call[1][0]) > 0)

    @patch("jupyter_deploy.cmd_utils.run_cmd_and_capture_output")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_record_noop_when_both_flags_are_false(self, mock_variable_handler_cls: Mock, mock_capture: Mock) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        handler = TerraformConfigHandler(Path("/fake/path"), Mock())

        # Act
        handler.record()

        # Assert
        mock_capture.assert_not_called()
        mock_vars_fns["sync_project_variables_config"].assert_not_called()

    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.engine.terraform.tf_plan.format_plan_variables")
    @patch("jupyter_deploy.engine.terraform.tf_plan.extract_variables_from_json_plan")
    @patch("jupyter_deploy.cmd_utils.run_cmd_and_capture_output")
    @patch("rich.console.Console")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_record_calls_plan_and_save_vars_only_on_var_flag(
        self,
        mock_variable_handler_cls: Mock,
        mock_console: Mock,
        mock_capture: Mock,
        mock_extract: Mock,
        mock_format: Mock,
        mock_write: Mock,
    ) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        mock_console_instance = Mock()
        mock_console.return_value = mock_console_instance
        mock_capture.return_value = "i-am-a-serialized-plan"

        mock_var1 = Mock()
        mock_var2 = Mock()
        mock_secret1 = Mock()
        mock_var1.value = 1
        mock_var2.value = "two"
        mock_secret1.value = "nuclear-codes"
        mock_extract.return_value = ({"var1": mock_var1, "var2": mock_var2}, {"secret1": mock_secret1})
        mock_format.return_value = ["var1 = 1\n", 'var2 = "two"\n']

        path = Path("/fake/path")
        handler = TerraformConfigHandler(path, Mock())

        # Act
        handler.record(record_vars=True)

        # Assert
        mock_capture.assert_called_once()
        mock_plan_cmd_capture = mock_capture.mock_calls[0]
        self.assertEqual(mock_plan_cmd_capture[1][0][:3], ["terraform", "show", "-json"])

        mock_extract.assert_called_once_with("i-am-a-serialized-plan")
        mock_format.assert_called_once_with({"var1": mock_var1, "var2": mock_var2})
        mock_write.assert_called_once()

        mock_write_call = mock_write.mock_calls[0]
        self.assertEqual(mock_write_call[1][0], TestTerraformConfigHandler.MOCK_RECORD_VARS_PATH)
        self.assertIn("var1 = 1\n", mock_write_call[1][1])
        self.assertIn('var2 = "two"\n', mock_write_call[1][1])

        mock_vars_fns["sync_project_variables_config"].assert_called_once_with({"var1": 1, "var2": "two"})

    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.engine.terraform.tf_plan.format_plan_variables")
    @patch("jupyter_deploy.engine.terraform.tf_plan.extract_variables_from_json_plan")
    @patch("jupyter_deploy.cmd_utils.run_cmd_and_capture_output")
    @patch("rich.console.Console")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_record_calls_plan_and_save_secrets_only_on_secret_flag(
        self,
        mock_variable_handler_cls: Mock,
        mock_console: Mock,
        mock_capture: Mock,
        mock_extract: Mock,
        mock_format: Mock,
        mock_write: Mock,
    ) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        mock_console_instance = Mock()
        mock_console.return_value = mock_console_instance
        mock_capture.return_value = "i-am-a-serialized-plan"

        mock_var1 = Mock()
        mock_var2 = Mock()
        mock_secret = Mock()
        mock_var1.value = 1
        mock_var2.value = "two"
        mock_secret.value = "nuclear-codes"
        mock_extract.return_value = ({"var1": mock_var1, "var2": mock_var2}, {"secret1": mock_secret})
        mock_format.return_value = ['secret1 = "nuclear-codes"\n']

        path = Path("/fake/path")
        handler = TerraformConfigHandler(path, Mock())

        # Act
        handler.record(record_secrets=True)

        # Assert
        mock_capture.assert_called_once()
        mock_extract.assert_called_once_with("i-am-a-serialized-plan")
        mock_format.assert_called_once_with({"secret1": mock_secret})
        mock_write.assert_called_once()

        mock_write_call = mock_write.mock_calls[0]
        self.assertEqual(mock_write_call[1][0], TestTerraformConfigHandler.MOCK_RECORD_SECRETS_PATH)
        self.assertIn('secret1 = "nuclear-codes"\n', mock_write_call[1][1])

        mock_vars_fns["sync_project_variables_config"].assert_called_once_with({"secret1": "nuclear-codes"})

    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.engine.terraform.tf_plan.format_plan_variables")
    @patch("jupyter_deploy.engine.terraform.tf_plan.extract_variables_from_json_plan")
    @patch("jupyter_deploy.cmd_utils.run_cmd_and_capture_output")
    @patch("rich.console.Console")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_record_should_call_plan_only_once_when_both_flags_are_passed(
        self,
        mock_variable_handler_cls: Mock,
        mock_console: Mock,
        mock_capture: Mock,
        mock_extract: Mock,
        mock_format: Mock,
        mock_write: Mock,
    ) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        mock_console_instance = Mock()
        mock_console.return_value = mock_console_instance
        mock_print = Mock()
        mock_console_instance.print = mock_print
        mock_capture.return_value = "i-am-a-serialized-plan"
        mock_var1 = Mock()
        mock_var2 = Mock()
        mock_secret = Mock()
        mock_var1.value = 1
        mock_var2.value = "two"
        mock_secret.value = "nuclear-codes"
        mock_extract.return_value = ({"var1": mock_var1, "var2": mock_var2}, {"secret1": mock_secret})
        mock_format.side_effect = [["var1 = 1\n", 'var2 = "two"\n'], ['secret1 = "nuclear-codes"\n']]

        path = Path("/fake/path")
        handler = TerraformConfigHandler(path, Mock())

        # Act
        handler.record(record_vars=True, record_secrets=True)

        # Assert
        mock_capture.assert_called_once()
        mock_extract.assert_called_once_with("i-am-a-serialized-plan")
        self.assertEqual(mock_format.call_count, 2)
        self.assertEqual(mock_write.call_count, 2)

        mock_write_vars_call = mock_write.mock_calls[0]
        self.assertEqual(mock_write_vars_call[1][0], TestTerraformConfigHandler.MOCK_RECORD_VARS_PATH)
        self.assertIn("var1 = 1\n", mock_write_vars_call[1][1])
        self.assertIn('var2 = "two"\n', mock_write_vars_call[1][1])

        mock_write_secrets_call = mock_write.mock_calls[1]
        self.assertEqual(mock_write_secrets_call[1][0], TestTerraformConfigHandler.MOCK_RECORD_SECRETS_PATH)
        self.assertIn('secret1 = "nuclear-codes"\n', mock_write_secrets_call[1][1])

        mock_vars_fns["sync_project_variables_config"].assert_called_once_with(
            {"var1": 1, "var2": "two", "secret1": "nuclear-codes"}
        )

    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.cmd_utils.run_cmd_and_capture_output")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_catches_plan_retrieve_errors_and_print(
        self,
        mock_variable_handler_cls: Mock,
        mock_capture: Mock,
        mock_write: Mock,
    ) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        mock_capture.side_effect = subprocess.CalledProcessError(
            1, ["terraform", "show", "-json"], "something went wrong", None
        )
        handler = TerraformConfigHandler(Path("/fake/path"), Mock())

        # Act
        handler.record(record_vars=True)

        # Assert
        mock_write.assert_not_called()
        mock_vars_fns["sync_project_variables_config"].assert_not_called()

    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.engine.terraform.tf_plan.extract_variables_from_json_plan")
    @patch("jupyter_deploy.cmd_utils.run_cmd_and_capture_output")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_catches_plan_validation_errors(
        self,
        mock_variable_handler_cls: Mock,
        mock_capture: Mock,
        mock_extract: Mock,
        mock_write: Mock,
    ) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        mock_extract.side_effect = ValidationError("some error", [])
        handler = TerraformConfigHandler(Path("/fake/path"), Mock())

        # Act
        handler.record(record_secrets=True)

        # Assert
        mock_capture.assert_called_once()
        mock_write.assert_not_called()
        mock_vars_fns["sync_project_variables_config"].assert_not_called()

    @patch("jupyter_deploy.fs_utils.write_inline_file_content")
    @patch("jupyter_deploy.engine.terraform.tf_plan.extract_variables_from_json_plan")
    @patch("jupyter_deploy.cmd_utils.run_cmd_and_capture_output")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_catches_plan_json_parse_errors(
        self,
        mock_variable_handler_cls: Mock,
        mock_capture: Mock,
        mock_extract: Mock,
        mock_write: Mock,
    ) -> None:
        # Arrange
        mock_vars_handler, mock_vars_fns = self.get_mock_variable_handler_and_fns()
        mock_variable_handler_cls.return_value = mock_vars_handler
        mock_capture.return_value = "i-am-a-serialized-plan"
        mock_extract.side_effect = ValueError("Invalid JSON")
        handler = TerraformConfigHandler(Path("/fake/path"), Mock())

        # Act
        handler.record(record_secrets=True)

        # Assert
        mock_capture.assert_called_once()
        mock_extract.assert_called_once()
        mock_write.assert_not_called()
        mock_vars_fns["sync_project_variables_config"].assert_not_called()



import unittest
from unittest.mock import Mock, patch

from rich.console import Console

from jupyter_deploy.provider.aws.aws_ssm_runner import AwsSsmInstruction, AwsSsmRunner
from jupyter_deploy.provider.instruction_runner import InterruptInstructionError
from jupyter_deploy.provider.resolved_argdefs import (
    IntResolvedInstructionArgument,
    ListStrResolvedInstructionArgument,
    ResolvedInstructionArgument,
    StrResolvedInstructionArgument,
)


class TestAwsSsmRunner(unittest.TestCase):
    @patch("boto3.client")
    def test_aws_ssm_runner_instantiates_client(self, mock_boto3_client: Mock) -> None:
        # Arrange
        mock_client = Mock()
        mock_boto3_client.return_value = mock_client
        region_name = "us-west-2"

        # Act
        runner = AwsSsmRunner(region_name=region_name)

        # Assert
        mock_boto3_client.assert_called_once_with("ssm", region_name=region_name)
        self.assertEqual(runner.client, mock_client)

    def test_aws_ssm_raise_not_implemented_error_on_unmatched_instruction_name(self) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        invalid_instruction = "invalid-instruction"

        # Act & Assert
        with self.assertRaises(NotImplementedError) as context:
            runner.execute_instruction(instruction_name=invalid_instruction, resolved_arguments={}, console=console)

        self.assertIn(f"aws.ssm.{invalid_instruction}", str(context.exception))


class TestVerifyEc2InstanceAccessible(unittest.TestCase):
    @patch("jupyter_deploy.api.aws.ssm.ssm_session.describe_instance_information")
    def test_happy_case_calls_describe_return_true_no_console_print(self, mock_describe: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        instance_id = "i-1234567890abcdef0"

        # Setup mock to return online status
        mock_describe.return_value = {
            "InstanceId": instance_id,
            "PingStatus": "Online",
            "LastPingDateTime": "2023-01-01T00:00:00.000Z",
        }

        # Act
        result = runner._verify_ec2_instance_accessible(instance_id, console)

        # Assert
        self.assertTrue(result)
        mock_describe.assert_called_once_with(runner.client, instance_id=instance_id)
        # Console print should not be called when silent_success=True (default)
        console.print.assert_not_called()

    @patch("jupyter_deploy.api.aws.ssm.ssm_session.describe_instance_information")
    def test_happy_case_with_silent_success_false_print_something(self, mock_describe: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        instance_id = "i-1234567890abcdef0"

        # Setup mock to return online status
        mock_describe.return_value = {
            "InstanceId": instance_id,
            "PingStatus": "Online",
            "LastPingDateTime": "2023-01-01T00:00:00.000Z",
        }

        # Act
        result = runner._verify_ec2_instance_accessible(instance_id, console, silent_success=False)

        # Assert
        self.assertTrue(result)
        mock_describe.assert_called_once_with(runner.client, instance_id=instance_id)
        # Console print should be called when silent_success=False
        console.print.assert_called_once()
        self.assertIn(instance_id, console.print.mock_calls[0][1][0])

    @patch("jupyter_deploy.api.aws.ssm.ssm_session.describe_instance_information")
    def test_ping_status_connection_lost_return_false_and_print_irrespective_of_flag(self, mock_describe: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        instance_id = "i-1234567890abcdef0"
        last_ping_date = "2023-01-01T00:00:00.000Z"

        # Setup mock to return ConnectionLost status
        mock_describe.return_value = {
            "InstanceId": instance_id,
            "PingStatus": "ConnectionLost",
            "LastPingDateTime": last_ping_date,
        }

        # Test both with silent_success=True and False
        for silent_success in [True, False]:
            # Reset console mock between iterations
            console.reset_mock()

            # Act
            result = runner._verify_ec2_instance_accessible(instance_id, console, silent_success=silent_success)

            # Assert
            self.assertFalse(result)
            # Console print should be called regardless of silent_success for error cases
            console.print.assert_called_once()
            # Verify that the print indicates connection lost
            self.assertIn(instance_id, console.print.mock_calls[0][1][0])
            self.assertIn(last_ping_date, console.print.mock_calls[0][1][0])
            self.assertEqual(console.print.mock_calls[0][2].get("style"), "red")

    @patch("jupyter_deploy.api.aws.ssm.ssm_session.describe_instance_information")
    def test_ping_status_inactive_return_false_and_print_irrespective_of_flag(self, mock_describe: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        instance_id = "i-1234567890abcdef0"

        # Setup mock to return Inactive status
        mock_describe.return_value = {
            "InstanceId": instance_id,
            "PingStatus": "Inactive",
        }

        # Test both with silent_success=True and False
        for silent_success in [True, False]:
            # Reset console mock between iterations
            console.reset_mock()

            # Act
            result = runner._verify_ec2_instance_accessible(instance_id, console, silent_success=silent_success)

            # Assert
            self.assertFalse(result)
            # Console print should be called regardless of silent_success for error cases
            console.print.assert_called_once()
            self.assertIn(instance_id, console.print.mock_calls[0][1][0])
            self.assertEqual(console.print.mock_calls[0][2].get("style"), "red")

    @patch("jupyter_deploy.api.aws.ssm.ssm_session.describe_instance_information")
    def test_other_ping_status_return_false_and_print_irrespective_of_flag(self, mock_describe: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        instance_id = "i-1234567890abcdef0"

        # Setup mock to return unknown/empty status
        mock_describe.return_value = {
            "InstanceId": instance_id,
            # No PingStatus provided
        }

        # Test both with silent_success=True and False
        for silent_success in [True, False]:
            # Reset console mock between iterations
            console.reset_mock()

            # Act
            result = runner._verify_ec2_instance_accessible(instance_id, console, silent_success=silent_success)

            # Assert
            self.assertFalse(result)
            # Console print should be called regardless of silent_success for error cases
            console.print.assert_called_once()
            self.assertIn(instance_id, console.print.mock_calls[0][1][0])
            self.assertEqual(console.print.mock_calls[0][2].get("style"), "red")

    @patch("jupyter_deploy.api.aws.ssm.ssm_session.describe_instance_information")
    def test_bubbles_up_errors_from_api(self, mock_describe: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        instance_id = "i-1234567890abcdef0"

        # Setup mock to raise an exception
        mock_describe.side_effect = Exception("API Error")

        # Act & Assert
        with self.assertRaises(Exception) as context:
            runner._verify_ec2_instance_accessible(instance_id, console)

        self.assertEqual(str(context.exception), "API Error")


class TestSendCmdToOneInstanceAndWaitSync(unittest.TestCase):
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.AwsSsmRunner._verify_ec2_instance_accessible")
    @patch("jupyter_deploy.api.aws.ssm.ssm_command.send_cmd_to_one_instance_and_wait_sync")
    def test_execute_happy_path_without_parameters_or_optional_args(
        self, mock_send_cmd: Mock, mock_verify: Mock
    ) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        document_name = "some-doc-name"
        instance_id = "i-1234567890abcdef0"

        # Mock the verify method to return True (agent is accessible)
        mock_verify.return_value = True

        # Added standard error content to match the updated implementation
        mock_send_cmd.return_value = {
            "Status": "Success",
            "StandardOutputContent": "Command output",
            "StandardErrorContent": "Command error",
        }

        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "document_name": StrResolvedInstructionArgument(argument_name="document_name", value=document_name),
            "instance_id": StrResolvedInstructionArgument(argument_name="instance_id", value=instance_id),
        }

        # Act
        result = runner.execute_instruction(
            instruction_name=AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC,
            resolved_arguments=resolved_arguments,
            console=console,
        )

        # Assert
        # Verify that SSM agent connection was checked
        mock_verify.assert_called_once_with(instance_id=instance_id, console=console)

        # Update to include default timeout values
        mock_send_cmd.assert_called_once_with(
            runner.client,
            document_name=document_name,
            instance_id=instance_id,
            timeout_seconds=30,  # Default value
            wait_after_send_seconds=2,  # Default value
        )

        self.assertEqual(result["Status"].value, "Success")
        self.assertEqual(result["StandardOutputContent"].value, "Command output")
        self.assertEqual(result["StandardErrorContent"].value, "Command error")
        # There should be at least one print call (for the execute info)
        self.assertGreaterEqual(console.print.call_count, 1)
        # Find the call with the document and instance info
        found_info_print = False
        for call in console.print.mock_calls:
            if (
                len(call[1]) > 0
                and isinstance(call[1][0], str)
                and document_name in call[1][0]
                and instance_id in call[1][0]
            ):
                found_info_print = True
                break
        self.assertTrue(found_info_print, "Did not find expected console print with document name and instance id")

    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.AwsSsmRunner._verify_ec2_instance_accessible")
    @patch("jupyter_deploy.api.aws.ssm.ssm_command.send_cmd_to_one_instance_and_wait_sync")
    def test_execute_happy_path_with_parameters(self, mock_send_cmd: Mock, mock_verify: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        document_name = "AWS-RunShellScript"
        instance_id = "i-1234567890abcdef0"
        commands = ["echo 'Hello World'", "ls -la"]
        workingDirectory = "/tmp"

        # Mock the verify method to return True (agent is accessible)
        mock_verify.return_value = True

        mock_send_cmd.return_value = {
            "Status": "Success",
            "StandardOutputContent": "Hello World\nfile1 file2",
        }

        # Setup arguments with custom parameters
        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "document_name": StrResolvedInstructionArgument(argument_name="document_name", value=document_name),
            "instance_id": StrResolvedInstructionArgument(argument_name="instance_id", value=instance_id),
            "commands": ListStrResolvedInstructionArgument(argument_name="commands", value=commands),
            "workingDirectory": StrResolvedInstructionArgument(
                argument_name="workingDirectory", value=workingDirectory
            ),
        }

        # Act
        result = runner.execute_instruction(
            instruction_name=AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC,
            resolved_arguments=resolved_arguments,
            console=console,
        )

        # Assert
        # Verify that SSM agent connection was checked
        mock_verify.assert_called_once_with(instance_id=instance_id, console=console)

        # Check that the parameters were passed correctly
        mock_send_cmd.assert_called_once()
        call_args = mock_send_cmd.call_args[0]
        call_kwargs = mock_send_cmd.call_args[1]

        self.assertEqual(call_args[0], runner.client)
        self.assertEqual(call_kwargs["document_name"], document_name)
        self.assertEqual(call_kwargs["instance_id"], instance_id)
        self.assertEqual(call_kwargs["timeout_seconds"], 30)  # Default value
        self.assertEqual(call_kwargs["wait_after_send_seconds"], 2)  # Default value
        self.assertEqual(call_kwargs["commands"], commands)  # Custom parameter
        # The implementation converts string parameters to a list
        self.assertEqual(call_kwargs["workingDirectory"], [workingDirectory])  # Custom parameter

        self.assertEqual(result["Status"].value, "Success")
        self.assertEqual(result["StandardOutputContent"].value, "Hello World\nfile1 file2")

        # Check that console shows parameters
        self.assertGreaterEqual(console.print.call_count, 1)

        # Find the call with parameter info
        found_params_print = False
        for call in console.print.mock_calls:
            if len(call[1]) > 0 and isinstance(call[1][0], str) and "parameters" in call[1][0]:
                found_params_print = True
                self.assertIn(document_name, call[1][0])
                self.assertIn(instance_id, call[1][0])
                break
        self.assertTrue(found_params_print, "Did not find expected console print with parameters")

    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.AwsSsmRunner._verify_ec2_instance_accessible")
    @patch("jupyter_deploy.api.aws.ssm.ssm_command.send_cmd_to_one_instance_and_wait_sync")
    def test_execute_happy_path_with_optional_args(self, mock_send_cmd: Mock, mock_verify: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        document_name = "AWS-RunShellScript"
        instance_id = "i-1234567890abcdef0"
        # Custom timeout values
        timeout_seconds = 120
        wait_after_send_seconds = 5

        # Mock the verify method to return True (agent is accessible)
        mock_verify.return_value = True

        mock_send_cmd.return_value = {
            "Status": "Success",
            "StandardOutputContent": "Command output",
        }

        # Setup arguments with optional arguments
        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "document_name": StrResolvedInstructionArgument(argument_name="document_name", value=document_name),
            "instance_id": StrResolvedInstructionArgument(argument_name="instance_id", value=instance_id),
            "timeout_seconds": IntResolvedInstructionArgument(argument_name="timeout_seconds", value=timeout_seconds),
            "wait_after_send_seconds": IntResolvedInstructionArgument(
                argument_name="wait_after_send_seconds", value=wait_after_send_seconds
            ),
        }

        # Act
        result = runner.execute_instruction(
            instruction_name=AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC,
            resolved_arguments=resolved_arguments,
            console=console,
        )

        # Assert
        # Verify that SSM agent connection was checked
        mock_verify.assert_called_once_with(instance_id=instance_id, console=console)

        # Check that custom timeout values were used
        mock_send_cmd.assert_called_once_with(
            runner.client,
            document_name=document_name,
            instance_id=instance_id,
            timeout_seconds=timeout_seconds,  # Custom value
            wait_after_send_seconds=wait_after_send_seconds,  # Custom value
        )

        self.assertEqual(result["Status"].value, "Success")
        self.assertEqual(result["StandardOutputContent"].value, "Command output")

    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.AwsSsmRunner._verify_ec2_instance_accessible")
    @patch("jupyter_deploy.api.aws.ssm.ssm_command.send_cmd_to_one_instance_and_wait_sync")
    def test_execute_cmd_fail_prints_stdout_and_stderror(self, mock_send_cmd: Mock, mock_verify: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        document_name = "AWS-RunShellScript"
        instance_id = "i-1234567890abcdef0"

        # Mock the verify method to return True (agent is accessible)
        mock_verify.return_value = True

        # Setup mock to return failed status with stdout and stderr content
        mock_send_cmd.return_value = {
            "Status": "Failed",  # Failed status
            "StandardOutputContent": "Some output before failure  \n\n",
            "StandardErrorContent": "Error: Command not found \n",
        }

        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "document_name": StrResolvedInstructionArgument(argument_name="document_name", value=document_name),
            "instance_id": StrResolvedInstructionArgument(argument_name="instance_id", value=instance_id),
        }

        # Act
        result = runner.execute_instruction(
            instruction_name=AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC,
            resolved_arguments=resolved_arguments,
            console=console,
        )

        # Assert
        # Verify that SSM agent connection was checked
        mock_verify.assert_called_once_with(instance_id=instance_id, console=console)

        self.assertEqual(result["Status"].value, "Failed")
        self.assertEqual(result["StandardOutputContent"].value, "Some output before failure")

        # Verify that console.print was called with appropriate error messages
        # First call should be the executing message
        # Second call should indicate failure
        self.assertTrue(console.print.call_count >= 3)

        # Check for failure message
        failure_call_idx = -1
        for i, call in enumerate(console.print.mock_calls):
            if len(call[1]) > 0 and isinstance(call[1][0], str) and "failed" in call[1][0]:
                failure_call_idx = i
                break

        self.assertNotEqual(failure_call_idx, -1, "No failure message found in console output")
        failure_call = console.print.mock_calls[failure_call_idx]
        self.assertIn("failed", failure_call[1][0])
        self.assertEqual(failure_call[2].get("style"), "red")

        # Check that stderr content was printed somewhere
        stderr_found = False
        stdout_found = False

        for call in console.print.mock_calls:
            if len(call[1]) > 0 and call[1][0] == "Error: Command not found":
                stderr_found = True
            if len(call[1]) > 0 and call[1][0] == "Some output before failure":
                stdout_found = True

        self.assertTrue(stderr_found, "stderr content not found in console output")
        self.assertTrue(stdout_found, "stdout content not found in console output")

    def test_execute_raise_on_missing_or_invalid_type_instance_id(self) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        document_name = "AWS-RunShellScript"

        # Case 1: Missing instance_id
        resolved_arguments_missing: dict[str, ResolvedInstructionArgument] = {
            "document_name": StrResolvedInstructionArgument(argument_name="document_name", value=document_name)
        }

        # Act & Assert for missing instance_id
        with self.assertRaises(KeyError) as context:
            runner.execute_instruction(
                instruction_name=AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC,
                resolved_arguments=resolved_arguments_missing,
                console=console,
            )

        self.assertIn("instance_id", str(context.exception))

        # Case 2: Invalid type for instance_id
        resolved_arguments_invalid_type: dict[str, ResolvedInstructionArgument] = {
            "document_name": StrResolvedInstructionArgument(argument_name="document_name", value=document_name),
            "instance_id": ListStrResolvedInstructionArgument(
                argument_name="instance_id", value=["i-1234567890abcdef0"]
            ),
        }

        # Act & Assert for invalid type
        with self.assertRaises(TypeError):
            runner.execute_instruction(
                instruction_name=AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC,
                resolved_arguments=resolved_arguments_invalid_type,
                console=console,
            )

    def test_execute_raise_on_missing_or_invalid_type_doc_name(self) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        instance_id = "i-1234567890abcdef0"

        # Case 1: Missing document_name
        resolved_arguments_missing: dict[str, ResolvedInstructionArgument] = {
            "instance_id": StrResolvedInstructionArgument(argument_name="instance_id", value=instance_id)
        }

        # Act & Assert for missing document_name
        with self.assertRaises(KeyError) as context:
            runner.execute_instruction(
                instruction_name=AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC,
                resolved_arguments=resolved_arguments_missing,
                console=console,
            )

        self.assertIn("document_name", str(context.exception))

        # Case 2: Invalid type for document_name
        resolved_arguments_invalid_type: dict[str, ResolvedInstructionArgument] = {
            "document_name": ListStrResolvedInstructionArgument(argument_name="document_name", value=["doc-1"]),
            "instance_id": StrResolvedInstructionArgument(argument_name="instance_id", value=instance_id),
        }

        # Act & Assert for invalid type
        with self.assertRaises(TypeError):
            runner.execute_instruction(
                instruction_name=AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC,
                resolved_arguments=resolved_arguments_invalid_type,
                console=console,
            )

    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.AwsSsmRunner._verify_ec2_instance_accessible")
    @patch("jupyter_deploy.api.aws.ssm.ssm_command.send_cmd_to_one_instance_and_wait_sync")
    def test_execute_raise_when_api_handler_raise(self, mock_send_cmd: Mock, mock_verify: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        document_name = "AWS-RunShellScript"
        instance_id = "i-1234567890abcdef0"

        # Mock the verify method to return True (agent is accessible)
        mock_verify.return_value = True

        # Setup mock to raise an exception
        mock_send_cmd.side_effect = Exception("API Error")

        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "document_name": StrResolvedInstructionArgument(argument_name="document_name", value=document_name),
            "instance_id": StrResolvedInstructionArgument(argument_name="instance_id", value=instance_id),
        }

        # Act & Assert
        with self.assertRaises(Exception) as context:
            runner.execute_instruction(
                instruction_name=AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC,
                resolved_arguments=resolved_arguments,
                console=console,
            )

        self.assertEqual(str(context.exception), "API Error")
        # Verify that SSM agent connection was checked
        mock_verify.assert_called_once_with(instance_id=instance_id, console=console)

        mock_send_cmd.assert_called_once_with(
            runner.client,
            document_name=document_name,
            instance_id=instance_id,
            timeout_seconds=30,  # Default value
            wait_after_send_seconds=2,  # Default value
        )

    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.AwsSsmRunner._verify_ec2_instance_accessible")
    def test_execute_raise_when_verification_fails(self, mock_verify: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        document_name = "AWS-RunShellScript"
        instance_id = "i-1234567890abcdef0"

        # Mock the verify method to return False (agent is NOT accessible)
        mock_verify.return_value = False

        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "document_name": StrResolvedInstructionArgument(argument_name="document_name", value=document_name),
            "instance_id": StrResolvedInstructionArgument(argument_name="instance_id", value=instance_id),
        }

        # Act & Assert
        with self.assertRaises(InterruptInstructionError):
            runner.execute_instruction(
                instruction_name=AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC,
                resolved_arguments=resolved_arguments,
                console=console,
            )

        # Verify that SSM agent connection was checked
        mock_verify.assert_called_once_with(instance_id=instance_id, console=console)


class TestStartSession(unittest.TestCase):
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.AwsSsmRunner._verify_ec2_instance_accessible")
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.verify_utils.verify_tools_installation")
    def test_happy_case_calls_verify_methods_start_session(
        self, mock_verify_tools: Mock, mock_verify_ec2: Mock, mock_run_cmd: Mock
    ) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        target_id = "i-1234567890abcdef0"

        # Mock successful verifications
        mock_verify_tools.return_value = True
        mock_verify_ec2.return_value = True

        # Mock successful command execution
        mock_run_cmd.return_value = (0, False)  # Return code, timeout flag

        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "target_id": StrResolvedInstructionArgument(argument_name="target_id", value=target_id),
        }

        # Act
        result = runner.execute_instruction(
            instruction_name=AwsSsmInstruction.START_SESSION,
            resolved_arguments=resolved_arguments,
            console=console,
        )

        # Assert
        # Verify that all required checks were performed
        mock_verify_tools.assert_called_once()
        mock_verify_ec2.assert_called_once_with(instance_id=target_id, console=console, silent_success=False)

        # Verify that the session command was executed with the correct target
        mock_run_cmd.assert_called_once()
        cmd_args = mock_run_cmd.call_args[0][0]
        self.assertIn("--target", cmd_args)
        self.assertIn(target_id, cmd_args)

        # Result should be an empty dict for START_SESSION
        self.assertEqual(result, {})

        # Console should print informational messages
        console_messages = [call[1][0] for call in console.print.mock_calls if len(call[1]) > 0]
        self.assertTrue(any("Starting SSM session" in msg for msg in console_messages if isinstance(msg, str)))
        self.assertTrue(any("Type 'exit'" in msg for msg in console_messages if isinstance(msg, str)))

    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.verify_utils.verify_tools_installation")
    def test_stops_on_missing_tool_installations(self, mock_verify_tools: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        target_id = "i-1234567890abcdef0"

        # Mock failed tool verification
        mock_verify_tools.return_value = False

        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "target_id": StrResolvedInstructionArgument(argument_name="target_id", value=target_id),
        }

        # Act & Assert
        with self.assertRaises(InterruptInstructionError):
            runner.execute_instruction(
                instruction_name=AwsSsmInstruction.START_SESSION,
                resolved_arguments=resolved_arguments,
                console=console,
            )

        # Verify that the verification was called but no further processing happened
        mock_verify_tools.assert_called_once()

    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.AwsSsmRunner._verify_ec2_instance_accessible")
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.verify_utils.verify_tools_installation")
    def test_stops_on_ssm_agent_not_connected(self, mock_verify_tools: Mock, mock_verify_ec2: Mock) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        target_id = "i-1234567890abcdef0"

        # Mock successful tools verification but failed EC2 connection
        mock_verify_tools.return_value = True
        mock_verify_ec2.return_value = False

        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "target_id": StrResolvedInstructionArgument(argument_name="target_id", value=target_id),
        }

        # Act & Assert
        with self.assertRaises(InterruptInstructionError):
            runner.execute_instruction(
                instruction_name=AwsSsmInstruction.START_SESSION,
                resolved_arguments=resolved_arguments,
                console=console,
            )

        # Verify that both verifications were called but no further processing
        mock_verify_tools.assert_called_once()
        mock_verify_ec2.assert_called_once_with(instance_id=target_id, console=console, silent_success=False)

    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.AwsSsmRunner._verify_ec2_instance_accessible")
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.verify_utils.verify_tools_installation")
    def test_raises_on_subcommand_status_error(
        self, mock_verify_tools: Mock, mock_verify_ec2: Mock, mock_run_cmd: Mock
    ) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        target_id = "i-1234567890abcdef0"

        # Mock successful verifications
        mock_verify_tools.return_value = True
        mock_verify_ec2.return_value = True

        # Mock command execution with non-zero return code
        mock_run_cmd.return_value = (1, False)  # Return code 1 indicates error

        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "target_id": StrResolvedInstructionArgument(argument_name="target_id", value=target_id),
        }

        # Act & Assert
        with self.assertRaises(InterruptInstructionError):
            runner.execute_instruction(
                instruction_name=AwsSsmInstruction.START_SESSION,
                resolved_arguments=resolved_arguments,
                console=console,
            )

        # Verify that session command was executed but detected the error
        mock_run_cmd.assert_called_once()

    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.AwsSsmRunner._verify_ec2_instance_accessible")
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.verify_utils.verify_tools_installation")
    def test_raises_on_subcommand_timeout_error(
        self, mock_verify_tools: Mock, mock_verify_ec2: Mock, mock_run_cmd: Mock
    ) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        target_id = "i-1234567890abcdef0"

        # Mock successful verifications
        mock_verify_tools.return_value = True
        mock_verify_ec2.return_value = True

        # Mock command execution with timeout
        mock_run_cmd.return_value = (0, True)  # Return code 0 but timeout=True

        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "target_id": StrResolvedInstructionArgument(argument_name="target_id", value=target_id),
        }

        # Act & Assert
        with self.assertRaises(InterruptInstructionError):
            runner.execute_instruction(
                instruction_name=AwsSsmInstruction.START_SESSION,
                resolved_arguments=resolved_arguments,
                console=console,
            )

        # Verify that session command was executed but detected timeout
        mock_run_cmd.assert_called_once()


class TestExecuteInstructions(unittest.TestCase):
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.AwsSsmRunner._verify_ec2_instance_accessible")
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.verify_utils.verify_tools_installation")
    @patch("jupyter_deploy.provider.aws.aws_ssm_runner.cmd_utils.run_cmd_and_pipe_to_terminal")
    @patch("jupyter_deploy.api.aws.ssm.ssm_command.send_cmd_to_one_instance_and_wait_sync")
    @patch("jupyter_deploy.api.aws.ssm.ssm_session.describe_instance_information")
    def test_all_ssm_instructions_implemented(
        self,
        mock_describe_info: Mock,
        mock_send_cmd: Mock,
        mock_run_cmd: Mock,
        mock_verify_tools: Mock,
        mock_verify_ec2: Mock,
    ) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)

        # Setup mocks for all possible instructions
        mock_verify_ec2.return_value = True
        mock_verify_tools.return_value = True
        mock_run_cmd.return_value = (0, False)  # return code, timeout

        mock_describe_info.return_value = {
            "PingStatus": "Online",
            "InstanceId": "i-12345",
        }

        mock_send_cmd.return_value = {
            "Status": "Success",
            "StandardOutputContent": "Command output",
        }

        # Verify each instruction in AwsSsmInstruction can be executed
        for instruction in AwsSsmInstruction:
            # Reset mocks between iterations
            mock_verify_ec2.reset_mock()
            mock_verify_tools.reset_mock()
            mock_run_cmd.reset_mock()
            mock_send_cmd.reset_mock()
            mock_describe_info.reset_mock()
            console.reset_mock()

            # Basic arguments that work for any instruction
            base_resolved_arguments: dict[str, ResolvedInstructionArgument] = {}

            if instruction == AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC:
                base_resolved_arguments = {
                    "document_name": StrResolvedInstructionArgument(argument_name="document_name", value="test-doc"),
                    "instance_id": StrResolvedInstructionArgument(argument_name="instance_id", value="i-12345"),
                }
            elif instruction == AwsSsmInstruction.START_SESSION:
                base_resolved_arguments = {
                    "target_id": StrResolvedInstructionArgument(argument_name="target_id", value="i-12345"),
                }
            else:
                raise NotImplementedError(f"Instruction {instruction} not implemented")

            # Each enum instruction should be implemented in the runner
            result = runner.execute_instruction(
                instruction_name=instruction, resolved_arguments=base_resolved_arguments, console=console
            )

            # Simple verification that the instruction was executed correctly
            if instruction == AwsSsmInstruction.SEND_CMD_AND_WAIT_SYNC:
                mock_send_cmd.assert_called_once()
                mock_verify_ec2.assert_called_once()
                self.assertEqual(result["Status"].value, "Success")
            elif instruction == AwsSsmInstruction.START_SESSION:
                mock_verify_tools.assert_called_once()
                mock_verify_ec2.assert_called_once()
                mock_run_cmd.assert_called_once()

    def test_raise_not_implemented_error_on_unrecognized_instruction(self) -> None:
        # Arrange
        runner = AwsSsmRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        invalid_instruction = "invalid-instruction"

        resolved_arguments: dict[str, ResolvedInstructionArgument] = {
            "document_name": StrResolvedInstructionArgument(argument_name="document_name", value="test-doc"),
            "instance_id": StrResolvedInstructionArgument(argument_name="instance_id", value="i-12345"),
        }

        # Act & Assert
        with self.assertRaises(NotImplementedError) as context:
            runner.execute_instruction(
                instruction_name=invalid_instruction, resolved_arguments=resolved_arguments, console=console
            )

        self.assertIn(f"aws.ssm.{invalid_instruction}", str(context.exception))



from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.engine.engine_variables import EngineVariablesHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform import tf_outputs, tf_variables
from jupyter_deploy.handlers.base_project_handler import BaseProjectHandler
from jupyter_deploy.provider import manifest_command_runner as cmd_runner
from jupyter_deploy.provider.resolved_clidefs import StrResolvedCliParameter


class OrganizationHandler(BaseProjectHandler):
    """Handler class to manage organization access to a jupyter app."""

    _output_handler: EngineOutputsHandler
    _variable_handler: EngineVariablesHandler

    def __init__(self) -> None:
        """Instantiate the Organization handler."""
        super().__init__()

        if self.engine == EngineType.TERRAFORM:
            self._output_handler = tf_outputs.TerraformOutputsHandler(
                project_path=self.project_path, project_manifest=self.project_manifest
            )
            self._variable_handler = tf_variables.TerraformVariablesHandler(
                project_path=self.project_path, project_manifest=self.project_manifest
            )
        else:
            raise NotImplementedError(f"OutputsHandler implementation not found for engine: {self.engine}")

    def set_organization(self, organization: str) -> None:
        """Allowlist the organization whose members or teams may access the Jupyter app."""
        command = self.project_manifest.get_command("organization.set")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        success, _ = runner.run_command_sequence(
            command,
            cli_paramdefs={
                "organization": StrResolvedCliParameter(parameter_name="organization", value=organization),
                "category": StrResolvedCliParameter(parameter_name="category", value="org"),
            },
        )
        if success:
            runner.update_variables(command)

    def unset_organization(self) -> None:
        """Remove allowlisting by organization to the Jupyter app."""
        command = self.project_manifest.get_command("organization.unset")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        success, _ = runner.run_command_sequence(
            command,
            cli_paramdefs={
                "category": StrResolvedCliParameter(parameter_name="category", value="org"),
            },
        )
        if success:
            runner.update_variables(command)

    def get_organization(self) -> str:
        """Return the organization allowlisted to access the Jupyter app."""
        command = self.project_manifest.get_command("organization.get")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )

        success, _ = runner.run_command_sequence(
            command,
            cli_paramdefs={
                "category": StrResolvedCliParameter(parameter_name="category", value="org"),
            },
        )
        if success:
            # No variables to update for get command
            return runner.get_result_value(command, "organization.get", str)
        return ""



import unittest
from pathlib import Path
from unittest.mock import Mock, patch

from jupyter_deploy.handlers.project.down_handler import DownHandler
from jupyter_deploy.manifest import JupyterDeployManifestV1


class TestDownHandler(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_manifest = JupyterDeployManifestV1(
            **{  # type: ignore
                "schema_version": 1,
                "template": {
                    "name": "mock-template-name",
                    "engine": "terraform",
                    "version": "1.0.0",
                },
            }
        )

    @patch("jupyter_deploy.engine.terraform.tf_down.TerraformDownHandler")
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_init_creates_terraform_handler(
        self, mock_cwd: Mock, mock_retrieve_manifest: Mock, mock_tf_handler_cls: Mock
    ) -> None:
        mock_cwd.return_value = Path("/mock/cwd")
        mock_retrieve_manifest.return_value = self.mock_manifest
        mock_tf_handler = Mock()
        mock_tf_handler_cls.return_value = mock_tf_handler

        handler = DownHandler()

        mock_tf_handler_cls.assert_called_once_with(project_path=Path("/mock/cwd"), project_manifest=self.mock_manifest)
        self.assertEqual(handler._handler, mock_tf_handler)

    @patch("jupyter_deploy.engine.terraform.tf_down.TerraformDownHandler")
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_destroy_delegates_to_handler(
        self, mock_cwd: Mock, mock_retrieve_manifest: Mock, mock_tf_handler_cls: Mock
    ) -> None:
        mock_cwd.return_value = Path("/mock/cwd")
        mock_retrieve_manifest.return_value = self.mock_manifest
        mock_tf_handler = Mock()
        mock_tf_handler.destroy.return_value = True
        mock_tf_handler_cls.return_value = mock_tf_handler

        handler = DownHandler()
        handler.destroy()

        mock_tf_handler.destroy.assert_called_once()

    @patch("jupyter_deploy.engine.terraform.tf_down.TerraformDownHandler")
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_destroy_propagates_exceptions(
        self, mock_cwd: Mock, mock_retrieve_manifest: Mock, mock_tf_handler_cls: Mock
    ) -> None:
        mock_cwd.return_value = Path("/mock/cwd")
        mock_retrieve_manifest.return_value = self.mock_manifest
        mock_tf_handler = Mock()
        mock_tf_handler.destroy.side_effect = Exception("Destroy failed")
        mock_tf_handler_cls.return_value = mock_tf_handler

        handler = DownHandler()

        with self.assertRaises(Exception) as context:
            handler.destroy()

        self.assertEqual(str(context.exception), "Destroy failed")
        mock_tf_handler.destroy.assert_called_once()

    @patch("jupyter_deploy.engine.terraform.tf_down.TerraformDownHandler")
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_destroy_with_auto_approve(
        self, mock_cwd: Mock, mock_retrieve_manifest: Mock, mock_tf_handler_cls: Mock
    ) -> None:
        mock_cwd.return_value = Path("/mock/cwd")
        mock_retrieve_manifest.return_value = self.mock_manifest
        mock_tf_handler = Mock()
        mock_tf_handler_cls.return_value = mock_tf_handler

        handler = DownHandler()
        handler.destroy(True)

        mock_tf_handler.destroy.assert_called_once_with(True)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_init_raises_value_error_for_unsupported_engine(self, mock_cwd: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_cwd.return_value = Path("/mock/cwd")
        mock_manifest = self.mock_manifest
        mock_manifest.template.engine = "UNSUPPORTED_ENGINE"  # type: ignore
        mock_retrieve_manifest.return_value = mock_manifest

        with self.assertRaises(ValueError):
            DownHandler()



import unittest
from pathlib import Path
from unittest import mock
from unittest.mock import Mock, patch

import yaml

from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.handlers.resource.server_handler import ServerHandler
from jupyter_deploy.manifest import JupyterDeployManifest, JupyterDeployManifestV1


class TestServerHandler(unittest.TestCase):
    mock_full_manifest: JupyterDeployManifest

    @classmethod
    def setUpClass(cls) -> None:
        full_manifest_path = Path(__file__).parent.parent.parent / "mock_manifest.yaml"
        with open(full_manifest_path) as f:
            manifest_content = f.read()
        manifest_parsed_content = yaml.safe_load(manifest_content)
        cls.mock_full_manifest = JupyterDeployManifestV1(
            **manifest_parsed_content  # type: ignore
        )

    def get_mock_manifest_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return mock manifest with functions defined as mock."""
        mock_manifest = Mock()
        mock_get_engine = Mock()
        mock_get_command = Mock()
        mock_get_validated_service = Mock()
        mock_get_engine.return_value = EngineType.TERRAFORM
        mock_get_command.return_value = Mock()
        mock_get_validated_service.return_value = "jupyter"
        mock_manifest.get_command = mock_get_command
        mock_manifest.get_engine = mock_get_engine
        mock_manifest.get_validated_service = mock_get_validated_service
        return mock_manifest, {
            "get_command": mock_get_command,
            "get_engine": mock_get_engine,
            "get_validated_service": mock_get_validated_service,
        }

    def get_mock_outputs_handler_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return mock output handler with functions defined as mock."""
        mock_output_handler = Mock()
        return mock_output_handler, {}

    def get_mock_manifest_cmd_runner_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return mock manifest cmd runner with functions defined as mock."""
        mock_cmd_runner_handler = Mock()
        mock_run_command_sequence = Mock()
        mock_get_result_value = Mock()

        mock_cmd_runner_handler.run_command_sequence = mock_run_command_sequence
        mock_cmd_runner_handler.get_result_value = mock_get_result_value

        mock_get_result_value.return_value = "IN_SERVICE"

        return mock_cmd_runner_handler, {
            "run_command_sequence": mock_run_command_sequence,
            "get_result_value": mock_get_result_value,
        }

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("pathlib.Path.cwd")
    def test_can_instantiate_terraform_project(
        self,
        mock_cwd: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        path = Path("/some/cur/dir")
        mock_cwd.return_value = path
        mock_output_handler = self.get_mock_outputs_handler_and_fns()[0]
        mock_tf_outputs_handler.return_value = mock_output_handler

        mock_variable_handler = Mock()
        mock_tf_variables_handler.return_value = mock_variable_handler

        mock_manifest, _ = self.get_mock_manifest_and_fns()
        mock_retrieve_manifest.return_value = mock_manifest

        handler = ServerHandler()

        mock_retrieve_manifest.assert_called_once()
        mock_tf_outputs_handler.assert_called_once_with(project_path=path, project_manifest=mock_manifest)
        mock_tf_variables_handler.assert_called_once_with(project_path=path, project_manifest=mock_manifest)

        self.assertEqual(handler._output_handler, mock_output_handler)
        self.assertEqual(handler._variable_handler, mock_variable_handler)
        self.assertEqual(handler.engine, EngineType.TERRAFORM)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_server_methods_raise_not_implemented_error_if_manifest_does_not_define_cmd(
        self, mock_tf_variables_handler: Mock, mock_tf_outputs_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_tf_outputs_handler.return_value = self.get_mock_outputs_handler_and_fns()[0]
        mock_tf_variables_handler.return_value = Mock()

        # Create a manifest with no commands defined
        no_cmd_manifest = JupyterDeployManifestV1(
            **{  # type: ignore
                "schema_version": 1,
                "template": {
                    "name": "mock-template-name",
                    "engine": "terraform",
                    "version": "1.0.0",
                },
            }
        )
        mock_retrieve_manifest.return_value = no_cmd_manifest

        handler = ServerHandler()

        with self.assertRaises(NotImplementedError):
            handler.get_server_status()

        with self.assertRaises(NotImplementedError):
            handler.start_server("all")

        with self.assertRaises(NotImplementedError):
            handler.stop_server("jupyter")

        with self.assertRaises(NotImplementedError):
            handler.restart_server("all")

        with self.assertRaises(NotImplementedError):
            handler.get_server_logs("traefik", [])

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    def test_server_methods_run_against_actual_manifest(
        self,
        mock_cmd_runner_class: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_full_manifest
        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner
        mock_tf_outputs_handler.return_value = self.get_mock_outputs_handler_and_fns()[0]
        mock_tf_variables_handler.return_value = Mock()

        handler = ServerHandler()

        # Test get_server_status
        status = handler.get_server_status()
        self.assertEqual(status, "IN_SERVICE")
        mock_cmd_runner_fns["run_command_sequence"].assert_called_with(mock.ANY, cli_paramdefs={})
        mock_cmd_runner_fns["get_result_value"].assert_called_with(mock.ANY, "server.status", str)

        # Test start_server
        mock_cmd_runner_fns["run_command_sequence"].reset_mock()
        mock_cmd_runner_fns["get_result_value"].reset_mock()
        handler.start_server("all")

        # Test stop_server
        mock_cmd_runner_fns["run_command_sequence"].reset_mock()
        handler.stop_server("jupyter")

        # Test restart_server
        mock_cmd_runner_fns["run_command_sequence"].reset_mock()
        handler.restart_server("traefik")

        # Test get_server_logs
        mock_cmd_runner_fns["run_command_sequence"].reset_mock()
        handler.get_server_logs("traefik", [])
        mock_cmd_runner_fns["get_result_value"].assert_any_call(mock.ANY, "server.logs", str)
        mock_cmd_runner_fns["get_result_value"].assert_any_call(mock.ANY, "server.errors", str)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    def test_server_methods_raise_if_run_command_raises(
        self,
        mock_cmd_runner_class: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        mock_retrieve_manifest.return_value = mock_manifest

        mock_tf_outputs_handler.return_value = self.get_mock_outputs_handler_and_fns()[0]
        mock_tf_variables_handler.return_value = Mock()

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner
        mock_cmd_runner_fns["run_command_sequence"].side_effect = RuntimeError()

        handler = ServerHandler()

        # verify methods raise
        with self.assertRaises(RuntimeError):
            handler.get_server_status()

        with self.assertRaises(RuntimeError):
            handler.start_server("all")

        with self.assertRaises(RuntimeError):
            handler.stop_server("jupyter")

        with self.assertRaises(RuntimeError):
            handler.restart_server("sidecars")

        with self.assertRaises(RuntimeError):
            handler.restart_server("oauth")

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    def test_get_status_method_raises_if_get_command_result_raises(
        self,
        mock_cmd_runner_class: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        mock_retrieve_manifest.return_value = mock_manifest

        mock_tf_outputs_handler.return_value = self.get_mock_outputs_handler_and_fns()[0]
        mock_tf_variables_handler.return_value = Mock()

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner
        mock_cmd_runner_fns["get_result_value"].side_effect = KeyError()

        handler = ServerHandler()

        # verify methods raise
        # add only commands that return a result here
        with self.assertRaises(KeyError):
            handler.get_server_status()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    def test_get_status_calls_run_command_and_return_result(
        self,
        mock_cmd_runner_class: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_manifest, mock_manifest_fns = self.get_mock_manifest_and_fns()
        mock_cmd = Mock()
        mock_manifest_fns["get_command"].return_value = mock_cmd

        mock_retrieve_manifest.return_value = mock_manifest
        mock_output_handler, _ = self.get_mock_outputs_handler_and_fns()

        mock_tf_outputs_handler.return_value = mock_output_handler
        mock_variable_handler = Mock()
        mock_tf_variables_handler.return_value = mock_variable_handler

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner

        # Act
        handler = ServerHandler()
        result = handler.get_server_status()

        # Verify
        self.assertEqual(result, "IN_SERVICE")
        mock_cmd_runner_class.assert_called_once()
        mock_manifest_fns["get_command"].assert_called_once_with("server.status")
        self.assertEqual(mock_cmd_runner_class.call_args[1]["output_handler"], mock_output_handler)
        self.assertEqual(mock_cmd_runner_class.call_args[1]["variable_handler"], mock_variable_handler)
        mock_cmd_runner_fns["run_command_sequence"].assert_called_once_with(mock_cmd, cli_paramdefs={})
        mock_cmd_runner_fns["get_result_value"].assert_called_once_with(mock_cmd, "server.status", str)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    def test_start_server_calls_run_command_with_correct_params(
        self,
        mock_cmd_runner_class: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_manifest, mock_manifest_fns = self.get_mock_manifest_and_fns()
        mock_cmd = Mock()
        mock_manifest_fns["get_command"].return_value = mock_cmd
        mock_manifest_fns["get_validated_service"].return_value = "all"

        mock_retrieve_manifest.return_value = mock_manifest
        mock_output_handler, _ = self.get_mock_outputs_handler_and_fns()

        mock_tf_outputs_handler.return_value = mock_output_handler
        mock_variable_handler = Mock()
        mock_tf_variables_handler.return_value = mock_variable_handler

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner

        # Act
        handler = ServerHandler()
        handler.start_server("all")

        # Verify
        mock_cmd_runner_class.assert_called_once()
        mock_manifest_fns["get_command"].assert_called_once_with("server.start")
        mock_manifest_fns["get_validated_service"].assert_called_once_with("all", allow_all=True)
        self.assertEqual(mock_cmd_runner_class.call_args[1]["output_handler"], mock_output_handler)
        self.assertEqual(mock_cmd_runner_class.call_args[1]["variable_handler"], mock_variable_handler)

        # Check that StrResolvedCliParameter objects are created correctly
        cli_paramdefs = mock_cmd_runner_fns["run_command_sequence"].call_args[1]["cli_paramdefs"]
        self.assertEqual(len(cli_paramdefs), 2)
        self.assertIn("action", cli_paramdefs)
        self.assertIn("service", cli_paramdefs)
        self.assertEqual(cli_paramdefs["action"].parameter_name, "action")
        self.assertEqual(cli_paramdefs["action"].value, "start")
        self.assertEqual(cli_paramdefs["service"].parameter_name, "service")
        self.assertEqual(cli_paramdefs["service"].value, "all")

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    def test_stop_server_calls_run_command_with_correct_params(
        self,
        mock_cmd_runner_class: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_manifest, mock_manifest_fns = self.get_mock_manifest_and_fns()
        mock_cmd = Mock()
        mock_manifest_fns["get_command"].return_value = mock_cmd
        mock_manifest_fns["get_validated_service"].return_value = "jupyter"

        mock_retrieve_manifest.return_value = mock_manifest
        mock_output_handler, _ = self.get_mock_outputs_handler_and_fns()

        mock_tf_outputs_handler.return_value = mock_output_handler
        mock_variable_handler = Mock()
        mock_tf_variables_handler.return_value = mock_variable_handler

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner

        # Act
        handler = ServerHandler()
        handler.stop_server("jupyter")

        # Verify
        mock_cmd_runner_class.assert_called_once()
        mock_manifest_fns["get_command"].assert_called_once_with("server.stop")
        mock_manifest_fns["get_validated_service"].assert_called_once_with("jupyter", allow_all=True)
        self.assertEqual(mock_cmd_runner_class.call_args[1]["output_handler"], mock_output_handler)
        self.assertEqual(mock_cmd_runner_class.call_args[1]["variable_handler"], mock_variable_handler)

        # Check CLI parameters
        cli_paramdefs = mock_cmd_runner_fns["run_command_sequence"].call_args[1]["cli_paramdefs"]
        self.assertEqual(len(cli_paramdefs), 2)
        self.assertIn("action", cli_paramdefs)
        self.assertIn("service", cli_paramdefs)
        self.assertEqual(cli_paramdefs["action"].parameter_name, "action")
        self.assertEqual(cli_paramdefs["action"].value, "stop")
        self.assertEqual(cli_paramdefs["service"].parameter_name, "service")
        self.assertEqual(cli_paramdefs["service"].value, "jupyter")

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    def test_restart_server_calls_run_command_with_correct_params(
        self,
        mock_cmd_runner_class: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_manifest, mock_manifest_fns = self.get_mock_manifest_and_fns()
        mock_cmd = Mock()
        mock_manifest_fns["get_command"].return_value = mock_cmd
        mock_manifest_fns["get_validated_service"].return_value = "sidecars"

        mock_retrieve_manifest.return_value = mock_manifest
        mock_output_handler, _ = self.get_mock_outputs_handler_and_fns()

        mock_tf_outputs_handler.return_value = mock_output_handler
        mock_variable_handler = Mock()
        mock_tf_variables_handler.return_value = mock_variable_handler

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner

        # Act
        handler = ServerHandler()
        handler.restart_server("sidecars")

        # Verify
        mock_cmd_runner_class.assert_called_once()
        mock_manifest_fns["get_command"].assert_called_once_with("server.restart")
        mock_manifest_fns["get_validated_service"].assert_called_once_with("sidecars", allow_all=True)
        self.assertEqual(mock_cmd_runner_class.call_args[1]["output_handler"], mock_output_handler)
        self.assertEqual(mock_cmd_runner_class.call_args[1]["variable_handler"], mock_variable_handler)

        # Check CLI parameters
        cli_paramdefs = mock_cmd_runner_fns["run_command_sequence"].call_args[1]["cli_paramdefs"]
        self.assertEqual(len(cli_paramdefs), 2)
        self.assertIn("action", cli_paramdefs)
        self.assertIn("service", cli_paramdefs)
        self.assertEqual(cli_paramdefs["action"].parameter_name, "action")
        self.assertEqual(cli_paramdefs["action"].value, "restart")
        self.assertEqual(cli_paramdefs["service"].parameter_name, "service")
        self.assertEqual(cli_paramdefs["service"].value, "sidecars")

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    def test_get_server_logs_calls_run_command_with_correct_params(
        self,
        mock_cmd_runner_class: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_manifest, mock_manifest_fns = self.get_mock_manifest_and_fns()
        mock_cmd = Mock()
        mock_manifest_fns["get_command"].return_value = mock_cmd
        mock_manifest_fns["get_validated_service"].return_value = "oauth"

        mock_retrieve_manifest.return_value = mock_manifest
        mock_output_handler, _ = self.get_mock_outputs_handler_and_fns()

        mock_tf_outputs_handler.return_value = mock_output_handler
        mock_variable_handler = Mock()
        mock_tf_variables_handler.return_value = mock_variable_handler

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_fns["get_result_value"].side_effect = ["some\nlogs", "some\nerrors"]
        mock_cmd_runner_class.return_value = mock_cmd_runner

        # Act
        handler = ServerHandler()
        logs, error_logs = handler.get_server_logs("oauth", ["-n", "200"])

        # Assert results
        self.assertEqual(logs, "some\nlogs")
        self.assertEqual(error_logs, "some\nerrors")

        # Verify
        mock_cmd_runner_class.assert_called_once()
        mock_manifest_fns["get_command"].assert_called_once_with("server.logs")
        mock_manifest_fns["get_validated_service"].assert_called_once_with("oauth", allow_all=False)
        self.assertEqual(mock_cmd_runner_fns["get_result_value"].call_count, 2)
        self.assertEqual(mock_cmd_runner_class.call_args[1]["output_handler"], mock_output_handler)
        self.assertEqual(mock_cmd_runner_class.call_args[1]["variable_handler"], mock_variable_handler)
        self.assertEqual(mock_cmd_runner_fns["get_result_value"].mock_calls[0][1][1], "server.logs")
        self.assertEqual(mock_cmd_runner_fns["get_result_value"].mock_calls[1][1][1], "server.errors")

        # Check CLI parameters
        cli_paramdefs = mock_cmd_runner_fns["run_command_sequence"].call_args[1]["cli_paramdefs"]
        self.assertEqual(len(cli_paramdefs), 2)
        self.assertIn("extra", cli_paramdefs)
        self.assertIn("service", cli_paramdefs)
        self.assertEqual(cli_paramdefs["extra"].parameter_name, "extra")
        self.assertEqual(cli_paramdefs["extra"].value, "-n 200")
        self.assertEqual(cli_paramdefs["service"].parameter_name, "service")
        self.assertEqual(cli_paramdefs["service"].value, "oauth")



import unittest
from typing import Any
from unittest.mock import Mock, patch

import parameterized  # type: ignore
import typer

from jupyter_deploy.engine.vardefs import (
    AnyNumericTemplateVariableDefinition,
    BoolTemplateVariableDefinition,
    DictStrTemplateVariableDefinition,
    FloatTemplateVariableDefinition,
    IntTemplateVariableDefinition,
    ListMapStrTemplateVariableDefinition,
    ListStrTemplateVariableDefinition,
    StrTemplateVariableDefinition,
    TemplateVariableDefinition,
)


class TestTemplateVariableClasses(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            ("simple_name", "simple-name"),
            ("camelCase", "camel-case"),
            ("snake_case", "snake-case"),
            ("PascalCase", "pascal-case"),
            ("with-dashes", "with-dashes"),
            ("with_MIXED_Case", "with-mixed-case"),
        ]
    )
    def test_should_convert_var_name_with_get_cli_var_name(self, input_name: str, expected_output: str) -> None:
        # Setup
        var_def: TemplateVariableDefinition = TemplateVariableDefinition(
            variable_name=input_name, description="Test description"
        )

        # Execute
        result = var_def.get_cli_var_name()

        # Assert
        self.assertEqual(result, expected_output)

    def test_should_add_non_none_default_as_preset_to_description(self) -> None:
        # Setup
        var_def = TemplateVariableDefinition(
            variable_name="test_var", description="Test description", default="default_value"
        )

        # Execute
        result = var_def.get_cli_description()

        # Assert
        self.assertIn("[preset: default_value]", result)
        self.assertIn("Test description", result)

    def test_should_not_add_preset_to_description_when_default_is_none(self) -> None:
        # Setup
        var_def: TemplateVariableDefinition = TemplateVariableDefinition(
            variable_name="test_var", description="Test description", default=None
        )

        # Execute
        result = var_def.get_cli_description()

        # Assert
        self.assertEqual(result, "Test description")
        self.assertNotIn("[preset:", result)

    def test_list_and_map_types_should_not_add_non_none_default_as_preset_to_description(self) -> None:
        # Setup
        var_def1 = ListStrTemplateVariableDefinition(
            variable_name="list_var", description="List description", default=["a", "b"]
        )
        var_def2 = DictStrTemplateVariableDefinition(
            variable_name="map_var", description="Map description", default={"Key": "Value"}
        )
        var_def3 = ListMapStrTemplateVariableDefinition(
            variable_name="list_map_var", description="List of map description", default=[{"key": "value"}]
        )

        # Execute
        result1 = var_def1.get_cli_description()
        result2 = var_def2.get_cli_description()
        result3 = var_def3.get_cli_description()

        # Assert
        self.assertNotIn("[preset: ", result1)
        self.assertIn("List description", result1)
        self.assertIn("--list-var", result1)

        self.assertNotIn("[preset: ", result2)
        self.assertIn("Map description", result2)
        self.assertIn("--map-var", result2)

        self.assertNotIn("[preset: ", result3)
        self.assertIn("List of map description", result3)
        self.assertIn("--list-map-var", result3)

    @patch("jupyter_deploy.engine.vardefs.str_utils.get_trimmed_header")
    def test_description_should_call_str_util_to_trim(self, mock_get_trimmed_header: Mock) -> None:
        # Setup
        mock_get_trimmed_header.return_value = "Trimmed description"
        var_def: TemplateVariableDefinition = TemplateVariableDefinition(
            variable_name="test_var", description="Test description"
        )

        # Execute
        var_def.get_cli_description()

        # Assert
        mock_get_trimmed_header.assert_called_once_with("Test description")

    def test_can_set_sensitive_bool_attribute(self) -> None:
        # Setup & Execute
        var_def_sensitive: TemplateVariableDefinition = TemplateVariableDefinition(
            variable_name="test_var", description="Test description", sensitive=True
        )

        var_def_not_sensitive: TemplateVariableDefinition = TemplateVariableDefinition(
            variable_name="test_var", description="Test description", sensitive=False
        )

        # Assert
        self.assertTrue(var_def_sensitive.sensitive)
        self.assertFalse(var_def_not_sensitive.sensitive)

    def test_should_instantiate_a_str_template_var_instance(self) -> None:
        # Setup
        var_def = StrTemplateVariableDefinition(
            variable_name="test_var",
            description="Test description",
            default="default_value",
            assigned_value="assigned_value",
        )

        # Assert
        self.assertEqual(var_def.variable_name, "test_var")
        self.assertEqual(var_def.description, "Test description")
        self.assertEqual(var_def.default, "default_value")
        self.assertEqual(var_def.assigned_value, "assigned_value")
        self.assertEqual(StrTemplateVariableDefinition.get_type(), str)

    def test_should_instantiate_an_int_template_var_instance(self) -> None:
        # Setup
        var_def = IntTemplateVariableDefinition(
            variable_name="test_var", description="Test description", default=42, assigned_value=99
        )

        # Assert
        self.assertEqual(var_def.variable_name, "test_var")
        self.assertEqual(var_def.description, "Test description")
        self.assertEqual(var_def.default, 42)
        self.assertEqual(var_def.assigned_value, 99)
        self.assertEqual(IntTemplateVariableDefinition.get_type(), int)

    def test_should_instantiate_a_float_template_var_instance(self) -> None:
        # Setup
        var_def = FloatTemplateVariableDefinition(
            variable_name="test_var", description="Test description", default=3.14, assigned_value=2.71
        )

        # Assert
        self.assertEqual(var_def.variable_name, "test_var")
        self.assertEqual(var_def.description, "Test description")
        self.assertEqual(var_def.default, 3.14)
        self.assertEqual(var_def.assigned_value, 2.71)
        self.assertEqual(FloatTemplateVariableDefinition.get_type(), float)

    def test_should_instantiate_a_bool_template_var_instance(self) -> None:
        # Setup
        var_def = BoolTemplateVariableDefinition(
            variable_name="test_var", description="Test description", default=True, assigned_value=False
        )

        # Assert
        self.assertEqual(var_def.variable_name, "test_var")
        self.assertEqual(var_def.description, "Test description")
        self.assertEqual(var_def.default, True)
        self.assertEqual(var_def.assigned_value, False)
        self.assertEqual(BoolTemplateVariableDefinition.get_type(), bool)

    def test_should_instantiate_a_anynumeric_template_var_instance(self) -> None:
        # Setup - with int values
        var_def_int = AnyNumericTemplateVariableDefinition(
            variable_name="test_var_int", description="Test description", default=42, assigned_value=99
        )

        # Setup - with float values
        var_def_float = AnyNumericTemplateVariableDefinition(
            variable_name="test_var_float", description="Test description", default=3.14, assigned_value=2.71
        )

        # Assert int values
        self.assertEqual(var_def_int.variable_name, "test_var_int")
        self.assertEqual(var_def_int.description, "Test description")
        self.assertEqual(var_def_int.default, 42)
        self.assertEqual(var_def_int.assigned_value, 99)

        # Assert float values
        self.assertEqual(var_def_float.variable_name, "test_var_float")
        self.assertEqual(var_def_float.description, "Test description")
        self.assertEqual(var_def_float.default, 3.14)
        self.assertEqual(var_def_float.assigned_value, 2.71)

    def test_should_instantiate_a_list_str_template_var_instance(self) -> None:
        # Setup
        default_list = ["item1", "item2"]
        assigned_list = ["item3", "item4"]
        var_def = ListStrTemplateVariableDefinition(
            variable_name="test_var", description="Test description", default=default_list, assigned_value=assigned_list
        )

        # Assert
        self.assertEqual(var_def.variable_name, "test_var")
        self.assertEqual(var_def.description, "Test description")
        self.assertEqual(var_def.default, default_list)
        self.assertEqual(var_def.assigned_value, assigned_list)
        self.assertEqual(ListStrTemplateVariableDefinition.get_type(), list[str])

    def test_should_instantiate_a_dict_str_template_var_instance(self) -> None:
        # Setup
        default_dict = {"key1": "value1", "key2": "value2"}
        assigned_dict = {"key3": "value3", "key4": "value4"}
        var_def = DictStrTemplateVariableDefinition(
            variable_name="test_var", description="Test description", default=default_dict, assigned_value=assigned_dict
        )

        # Assert
        self.assertEqual(var_def.variable_name, "test_var")
        self.assertEqual(var_def.description, "Test description")
        self.assertEqual(var_def.default, default_dict)
        self.assertEqual(var_def.assigned_value, assigned_dict)

        # with a fake type for typer
        self.assertEqual(DictStrTemplateVariableDefinition.get_type(), list[str])

    def test_should_instantiate_a_list_dict_str_template_var_instance(self) -> None:
        # Setup
        default_value = [{"key1": "val1"}, {"key1": "value1", "key2": "value2"}]
        assigned_value = [{"key1": "val1"}, {"key3": "value3", "key4": "value4"}]
        var_def = ListMapStrTemplateVariableDefinition(
            variable_name="test_var",
            description="Test description",
            default=default_value,
            assigned_value=assigned_value,
        )

        # Assert
        self.assertEqual(var_def.variable_name, "test_var")
        self.assertEqual(var_def.description, "Test description")
        self.assertEqual(var_def.default, default_value)
        self.assertEqual(var_def.assigned_value, assigned_value)

        # with a fake type for typer
        self.assertEqual(ListMapStrTemplateVariableDefinition.get_type(), list[str])

    @parameterized.parameterized.expand(
        [
            (StrTemplateVariableDefinition, True),
            (IntTemplateVariableDefinition, True),
            (FloatTemplateVariableDefinition, True),
            (BoolTemplateVariableDefinition, True),
            (ListStrTemplateVariableDefinition, True),
            (DictStrTemplateVariableDefinition, False),
            (ListMapStrTemplateVariableDefinition, False),
        ]
    )
    def test_get_validator_callback_is_set(self, clz: type, expect_none: bool) -> None:
        instance: TemplateVariableDefinition = clz(variable_name="var", description="desc")
        if expect_none:
            self.assertIsNone(instance.get_validator_callback())
        else:
            self.assertIsNotNone(instance.get_validator_callback())

    def test_list_map_str_validator_callback(self) -> None:
        instance = ListMapStrTemplateVariableDefinition(variable_name="var", description="desc")
        cb = instance.get_validator_callback()

        if not cb:
            raise ValueError("to keep mypy happy")

        self.assertEqual([], cb([]))
        self.assertEqual(["k1=v1,k2=v2", "k3=v3"], cb(["k1=v1,k2=v2", "k3=v3"]))

        with self.assertRaises(typer.BadParameter):
            cb({"key1": "key2"})
        with self.assertRaises(typer.BadParameter):
            cb(["key1:val1"])
        with self.assertRaises(typer.BadParameter):
            cb(["key1=val1", "key1val1"])
        with self.assertRaises(typer.BadParameter):
            cb(["key1=val1", ""])

    def test_map_str_validator_callback(self) -> None:
        instance = DictStrTemplateVariableDefinition(variable_name="var", description="desc")
        cb = instance.get_validator_callback()

        if not cb:
            raise ValueError("to keep mypy happy")

        self.assertEqual([], cb([]))
        self.assertEqual(["key1=val1", "key2=val2"], cb(["key1=val1", "key2=val2"]))

        with self.assertRaises(typer.BadParameter):
            cb({"key1": "key2"})
        with self.assertRaises(typer.BadParameter):
            cb(["key1:val1"])
        with self.assertRaises(typer.BadParameter):
            cb(["key1=val1", "key1val1"])
        with self.assertRaises(typer.BadParameter):
            cb(["key1=val1", ""])

    @parameterized.parameterized.expand(
        [
            (StrTemplateVariableDefinition, "val", "val"),
            (IntTemplateVariableDefinition, 2, 2),
            (FloatTemplateVariableDefinition, 3.1459, 3.1459),
            (BoolTemplateVariableDefinition, True, True),
            (ListStrTemplateVariableDefinition, ["a", "b"], ["a", "b"]),
            (DictStrTemplateVariableDefinition, ["key1=val1", "key2=val2"], {"key1": "val1", "key2": "val2"}),
            (
                ListMapStrTemplateVariableDefinition,
                ["key1=val1,key2=val2", "k1=v1,k2=v2"],
                [{"key1": "val1", "key2": "val2"}, {"k1": "v1", "k2": "v2"}],
            ),
        ]
    )
    def test_convert_assigned_value(self, clz: type, val: Any, expect: Any) -> None:
        instance: TemplateVariableDefinition = clz(variable_name="var", description="desc")
        converted_value = instance.convert_assigned_value(val)
        self.assertEqual(converted_value, expect)

    @parameterized.parameterized.expand(
        [
            (StrTemplateVariableDefinition, "val", 2),
            (IntTemplateVariableDefinition, 2, "two"),
            (FloatTemplateVariableDefinition, 3.1459, "pi"),
            (BoolTemplateVariableDefinition, True, 3),
            (ListStrTemplateVariableDefinition, ["a", "b"], {}),
            (DictStrTemplateVariableDefinition, {"key1": "val1", "key2": "val2"}, []),
            (ListMapStrTemplateVariableDefinition, [{"k1": "v1", "k2": "v2"}, {"k1": "v2"}], {}),
        ]
    )
    def test_validate_value(self, clz: type, valid_val: Any, invalid_val: Any) -> None:
        instance: TemplateVariableDefinition = clz(variable_name="var", description="desc")

        self.assertEqual(instance.validate_value(valid_val), valid_val)
        with self.assertRaises(TypeError):
            instance.validate_value(invalid_val)
        with self.assertRaises(ValueError):
            instance.validate_value(None)

    @parameterized.parameterized.expand(
        [
            (IntTemplateVariableDefinition, "2", 2),
            (FloatTemplateVariableDefinition, "3.1459", 3.1459),
            (BoolTemplateVariableDefinition, "true", True),
        ]
    )
    def test_validate_value_handles_conversion(self, clz: type, valid_val: Any, expect_val: Any) -> None:
        instance: TemplateVariableDefinition = clz(variable_name="var", description="desc")

        self.assertEqual(instance.validate_value(valid_val), expect_val)



import unittest
from typing import Any

from jupyter_deploy.engine.outdefs import StrTemplateOutputDefinition, TemplateOutputDefinition
from jupyter_deploy.provider.resolved_argdefs import (
    IntResolvedInstructionArgument,
    ListStrResolvedInstructionArgument,
    ResolvedInstructionArgument,
    StrResolvedInstructionArgument,
    require_arg,
    resolve_cliparam_argdef,
    resolve_output_argdef,
    resolve_result_argdef,
    retrieve_optional_arg,
)
from jupyter_deploy.provider.resolved_clidefs import (
    IntResolvedCliParameter,
    ListStrResolvedCliParameter,
    ResolvedCliParameter,
    StrResolvedCliParameter,
)
from jupyter_deploy.provider.resolved_resultdefs import (
    IntResolvedInstructionResult,
    ListStrResolvedInstructionResult,
    ResolvedInstructionResult,
    StrResolvedInstructionResult,
)


class CustomTemplateOutputDefinition(TemplateOutputDefinition):
    """Custom output definition for testing NotImplementedError."""

    def __init__(self, output_name: str, value: Any = None):
        super().__init__(output_name=output_name, value=value)


class CustomResolvedInstructionResult(ResolvedInstructionResult):
    """Custom result definition for testing NotImplementedError."""

    def __init__(self, result_name: str, value: Any = None):
        super().__init__(result_name=result_name, value=value)


class CustomResolvedCliParameter(ResolvedCliParameter):
    """Custom CLI parameter definition for testing NotImplementedError."""

    def __init__(self, parameter_name: str, value: Any = None):
        super().__init__(parameter_name=parameter_name, value=value)


class TestResolveOutputArgDef(unittest.TestCase):
    def test_resolves_existing_str_output(self) -> None:
        # Arrange
        outdefs: dict[str, TemplateOutputDefinition] = {
            "test_output": StrTemplateOutputDefinition(output_name="test_output", value="test_value")
        }

        # Act
        result = resolve_output_argdef(outdefs=outdefs, arg_name="test_arg", source_key="test_output")

        # Assert
        self.assertIsInstance(result, StrResolvedInstructionArgument)
        self.assertEqual(result.argument_name, "test_arg")
        self.assertEqual(result.value, "test_value")

    def test_raises_key_error_if_output_is_not_found(self) -> None:
        # Arrange
        outdefs: dict[str, TemplateOutputDefinition] = {
            "existing_output": StrTemplateOutputDefinition(output_name="existing_output", value="test_value")
        }

        # Act & Assert
        with self.assertRaises(KeyError) as context:
            resolve_output_argdef(outdefs, "test_arg", "non_existing_output")

        self.assertIn("non_existing_output", str(context.exception))

    def test_raises_not_implemented_error_if_type_does_not_match(self) -> None:
        # Arrange
        outdefs: dict[str, TemplateOutputDefinition] = {
            "custom_output": CustomTemplateOutputDefinition(output_name="custom_output", value="test_value")
        }

        # Act & Assert
        with self.assertRaises(NotImplementedError) as context:
            resolve_output_argdef(outdefs=outdefs, arg_name="test_arg", source_key="custom_output")

        self.assertIn(CustomTemplateOutputDefinition.__name__, str(context.exception))

    def test_raises_value_error_if_output_was_not_resolved(self) -> None:
        # Arrange
        outdefs: dict[str, TemplateOutputDefinition] = {
            "unresolved_output": StrTemplateOutputDefinition(output_name="unresolved_output", value=None)
        }

        # Act & Assert
        with self.assertRaises(ValueError) as context:
            resolve_output_argdef(outdefs, "test_arg", "unresolved_output")

        self.assertIn("unresolved_output", str(context.exception))


class TestResolveResultArgDef(unittest.TestCase):
    def test_resolves_all_result_types(self) -> None:
        # Arrange
        resultdefs: dict[str, ResolvedInstructionResult] = {
            "str_result": StrResolvedInstructionResult(result_name="str_result", value="test_value"),
            "int_result": IntResolvedInstructionResult(result_name="int_result", value=42),
            "list_str_result": ListStrResolvedInstructionResult(
                result_name="list_str_result", value=["value1", "value2"]
            ),
        }

        # Act & Assert for string result
        str_result = resolve_result_argdef(resultdefs=resultdefs, arg_name="str_arg", source_key="str_result")
        self.assertIsInstance(str_result, StrResolvedInstructionArgument)
        self.assertEqual(str_result.argument_name, "str_arg")
        self.assertEqual(str_result.value, "test_value")

        # Act & Assert for int result
        int_result = resolve_result_argdef(resultdefs=resultdefs, arg_name="int_arg", source_key="int_result")
        self.assertIsInstance(int_result, IntResolvedInstructionArgument)
        self.assertEqual(int_result.argument_name, "int_arg")
        self.assertEqual(int_result.value, 42)

        # Act & Assert for list of strings result
        list_str_result = resolve_result_argdef(
            resultdefs=resultdefs, arg_name="list_str_arg", source_key="list_str_result"
        )
        self.assertIsInstance(list_str_result, ListStrResolvedInstructionArgument)
        self.assertEqual(list_str_result.argument_name, "list_str_arg")
        self.assertEqual(list_str_result.value, ["value1", "value2"])

    def test_raises_key_error_if_result_is_not_found(self) -> None:
        # Arrange
        resultdefs: dict[str, ResolvedInstructionResult] = {
            "existing_result": StrResolvedInstructionResult(result_name="existing_result", value="test_value")
        }

        # Act & Assert
        with self.assertRaises(KeyError) as context:
            resolve_result_argdef(resultdefs, "test_arg", "non_existing_result")

        self.assertIn("non_existing_result", str(context.exception))

    def test_raises_not_implemented_error_if_type_does_not_match(self) -> None:
        # Arrange
        resultdefs: dict[str, ResolvedInstructionResult] = {
            "custom_result": CustomResolvedInstructionResult(result_name="custom_result", value="test_value")
        }

        # Act & Assert
        with self.assertRaises(NotImplementedError) as context:
            resolve_result_argdef(resultdefs=resultdefs, arg_name="test_arg", source_key="custom_result")

        self.assertIn(CustomResolvedInstructionResult.__name__, str(context.exception))


class TestResolveCliParamArgDef(unittest.TestCase):
    def test_resolves_all_cli_parameter_types(self) -> None:
        # Arrange
        paramdefs: dict[str, ResolvedCliParameter] = {
            "str_param": StrResolvedCliParameter(parameter_name="str_param", value="test_value"),
            "int_param": IntResolvedCliParameter(parameter_name="int_param", value=42),
            "list_str_param": ListStrResolvedCliParameter(parameter_name="list_str_param", value=["value1", "value2"]),
        }

        # Act & Assert for string parameter
        str_result = resolve_cliparam_argdef(paramdefs=paramdefs, arg_name="str_arg", source_key="str_param")
        self.assertIsInstance(str_result, StrResolvedInstructionArgument)
        self.assertEqual(str_result.argument_name, "str_arg")
        self.assertEqual(str_result.value, "test_value")

        # Act & Assert for int parameter
        int_result = resolve_cliparam_argdef(paramdefs=paramdefs, arg_name="int_arg", source_key="int_param")
        self.assertIsInstance(int_result, IntResolvedInstructionArgument)
        self.assertEqual(int_result.argument_name, "int_arg")
        self.assertEqual(int_result.value, 42)

        # Act & Assert for list of strings parameter
        list_str_result = resolve_cliparam_argdef(
            paramdefs=paramdefs, arg_name="list_str_arg", source_key="list_str_param"
        )
        self.assertIsInstance(list_str_result, ListStrResolvedInstructionArgument)
        self.assertEqual(list_str_result.argument_name, "list_str_arg")
        self.assertEqual(list_str_result.value, ["value1", "value2"])

    def test_raises_key_error_if_param_is_not_found(self) -> None:
        # Arrange
        paramdefs: dict[str, ResolvedCliParameter] = {
            "existing_param": StrResolvedCliParameter(parameter_name="existing_param", value="test_value")
        }

        # Act & Assert
        with self.assertRaises(KeyError) as context:
            resolve_cliparam_argdef(paramdefs, "test_arg", "non_existing_param")

        self.assertIn("non_existing_param", str(context.exception))

    def test_raises_not_implemented_error_if_type_does_not_match(self) -> None:
        # Arrange
        paramdefs: dict[str, ResolvedCliParameter] = {
            "custom_param": CustomResolvedCliParameter(parameter_name="custom_param", value="test_value")
        }

        # Act & Assert
        with self.assertRaises(NotImplementedError) as context:
            resolve_cliparam_argdef(paramdefs=paramdefs, arg_name="test_arg", source_key="custom_param")

        self.assertIn(CustomResolvedCliParameter.__name__, str(context.exception))


class TestRequireArg(unittest.TestCase):
    def test_return_when_found_and_type_matches(self) -> None:
        # Arrange
        str_arg = StrResolvedInstructionArgument(argument_name="str_arg", value="test_value")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"str_arg": str_arg}

        # Act
        result = require_arg(resolved_args, "str_arg", StrResolvedInstructionArgument)

        # Assert
        self.assertEqual(result, str_arg)

    def test_raises_key_error_if_not_found(self) -> None:
        # Arrange
        resolved_args: dict[str, ResolvedInstructionArgument] = {
            "existing_arg": StrResolvedInstructionArgument(argument_name="existing_arg", value="test_value")
        }

        # Act & Assert
        with self.assertRaises(KeyError) as context:
            require_arg(
                resolved_args=resolved_args, arg_name="non_existing_arg", arg_type=StrResolvedInstructionArgument
            )

        self.assertIn("non_existing_arg", str(context.exception))

    def test_raises_type_error_if_type_does_not_match(self) -> None:
        # Arrange
        str_arg = StrResolvedInstructionArgument(argument_name="str_arg", value="test_value")
        resolved_args_dict: dict[str, ResolvedInstructionArgument] = {"str_arg": str_arg}

        # Act & Assert
        with self.assertRaises(TypeError) as context:
            require_arg(resolved_args_dict, "str_arg", ListStrResolvedInstructionArgument)

        self.assertIn(
            ListStrResolvedInstructionArgument.__name__,
            str(context.exception),
        )
        self.assertIn(
            StrResolvedInstructionArgument.__name__,
            str(context.exception),
        )


class TestRetrieveOptionalArg(unittest.TestCase):
    def test_arg_found_return_value(self) -> None:
        # Arrange
        str_arg = StrResolvedInstructionArgument(argument_name="str_arg", value="test_value")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"str_arg": str_arg}

        # Act
        result = retrieve_optional_arg(
            resolved_args=resolved_args,
            arg_name="str_arg",
            arg_type=StrResolvedInstructionArgument,
            default_value="default_value",
        )

        # Assert
        self.assertEqual(result, str_arg)
        self.assertEqual(result.value, "test_value")

    def test_arg_not_found_return_default(self) -> None:
        # Arrange
        resolved_args: dict[str, ResolvedInstructionArgument] = {}

        # Act
        result = retrieve_optional_arg(
            resolved_args=resolved_args,
            arg_name="non_existing_arg",
            arg_type=StrResolvedInstructionArgument,
            default_value="default_value",
        )

        # Assert
        self.assertIsInstance(result, StrResolvedInstructionArgument)
        self.assertEqual(result.argument_name, "non_existing_arg")
        self.assertEqual(result.value, "default_value")

    def test_raises_if_found_but_type_does_not_match(self) -> None:
        # Arrange
        str_arg = StrResolvedInstructionArgument(argument_name="str_arg", value="test_value")
        resolved_args: dict[str, ResolvedInstructionArgument] = {"str_arg": str_arg}

        # Act & Assert
        with self.assertRaises(TypeError) as context:
            retrieve_optional_arg(
                resolved_args=resolved_args,
                arg_name="str_arg",
                arg_type=ListStrResolvedInstructionArgument,
                default_value=["default", "value"],
            )

        self.assertIn(
            ListStrResolvedInstructionArgument.__name__,
            str(context.exception),
        )
        self.assertIn(
            StrResolvedInstructionArgument.__name__,
            str(context.exception),
        )
